# -*- coding: utf-8 -*-
from datetime import datetime

from Acquisition import aq_base

from seantis.reservation.tests import IntegrationTestCase
from seantis.reservation import utils
from seantis.reservation.session import serialized
from seantis.reservation import exports
from seantis.reservation.export import ExportView, prepare_record


class TestExports(IntegrationTestCase):

    @serialized
    def test_reservations_export(self):
        self.login_manager()

        resource = self.create_resource()
        sc = resource.scheduler()

        start = datetime(2012, 2, 1, 12, 0)
        end = datetime(2012, 2, 1, 16, 0)
        some_date = datetime(2014, 1, 30, 13, 37)
        dates = (start, end)

        reservation_email = u'test@example.com'
        sc.allocate(dates, approve_manually=False, quota=2)[0]

        token1 = sc.reserve(
            reservation_email, dates,
            data=utils.mock_data_dictionary(
                {
                    'stop': u'hammertime!',
                    'bust': u'a move!',
                    'when': some_date
                }
            )
        )

        token2 = sc.reserve(
            reservation_email, dates,
            data=utils.mock_data_dictionary(
                {
                    'never': u'gonna',
                    'give': u'you up',
                    'when': some_date
                }
            )
        )

        dataset = exports.reservations.dataset(
            {resource.uuid(): resource}, 'en', 'all', 'all'
        )

        self.assertEqual(len(dataset), 2)

        existing_columns = 11
        token_1_unique = 2
        token_2_unique = 2
        token_common = 1

        self.assertEqual(
            len(dataset.headers),
            existing_columns + token_1_unique + token_2_unique + token_common
        )

        self.assertEqual(dataset.dict[0]['Token'], utils.string_uuid(token1))
        self.assertEqual(dataset.dict[0]['Mocktest.stop'], u'hammertime!')
        self.assertEqual(dataset.dict[0]['Mocktest.bust'], u'a move!')
        self.assertEqual(dataset.dict[0]['Mocktest.never'], None)
        self.assertEqual(dataset.dict[0]['Mocktest.give'], None)

        self.assertEqual(dataset.dict[1]['Token'], utils.string_uuid(token2))
        self.assertEqual(dataset.dict[1]['Mocktest.stop'], None)
        self.assertEqual(dataset.dict[1]['Mocktest.bust'], None)
        self.assertEqual(dataset.dict[1]['Mocktest.never'], u'gonna')
        self.assertEqual(dataset.dict[1]['Mocktest.give'], u'you up')

        self.assertEqual(dataset.dict[0]['Start'], start)
        self.assertEqual(dataset.dict[1]['Start'], start)
        self.assertEqual(dataset.dict[0]['End'], end)
        self.assertEqual(dataset.dict[1]['End'], end)

        self.assertEqual(dataset.dict[0]['Mocktest.when'], some_date)
        self.assertEqual(dataset.dict[1]['Mocktest.when'], some_date)

        # just make sure these don't raise exceptions
        for format in ('xls', 'xlsx', 'json', 'csv'):
            transform_record = lambda r: prepare_record(r, format)
            dataset = exports.reservations.dataset(
                {resource.uuid(): resource},
                'en', 'all', 'all', transform_record
            )
            getattr(dataset, format)

    @serialized
    def test_reservations_export_date_filter(self):
        self.login_manager()

        resource = self.create_resource()
        sc = resource.scheduler()

        this_year = (
            datetime(2013, 1, 1, 12, 0), datetime(2013, 1, 1, 13, 0)
        )

        next_year = (
            datetime(2014, 1, 1, 12, 0), datetime(2014, 1, 1, 13, 0)
        )

        reservation_email = u'test@example.com'
        sc.allocate(this_year, approve_manually=False)
        sc.allocate(next_year, approve_manually=False)

        sc.reserve(reservation_email, this_year)
        sc.reserve(reservation_email, next_year)

        dataset = exports.reservations.dataset(
            {resource.uuid(): resource}, 'en', year='all', month='all'
        )
        self.assertEqual(len(dataset), 2)

        dataset = exports.reservations.dataset(
            {resource.uuid(): resource}, 'en', year='all', month='1'
        )
        self.assertEqual(len(dataset), 2)

        dataset = exports.reservations.dataset(
            {resource.uuid(): resource}, 'en', year='2013', month='all'
        )
        self.assertEqual(len(dataset), 1)

        dataset = exports.reservations.dataset(
            {resource.uuid(): resource}, 'en', year='2010', month='all'
        )
        self.assertEqual(len(dataset), 0)

    @serialized
    def test_reservations_export_title(self):
        self.login_manager()

        resource = self.create_resource()
        sc = resource.scheduler()

        start = datetime(2012, 2, 1, 12, 0)
        end = datetime(2012, 2, 1, 16, 0)
        dates = (start, end)

        reservation_email = u'test@example.com'
        sc.allocate(dates, approve_manually=False, quota=2)[0]

        sc.reserve(reservation_email, dates)

        # with title
        resource.aq_inner.aq_parent.title = 'testtitel'

        dataset = exports.reservations.dataset(
            {resource.uuid(): resource}, 'en', 'all', 'all'
        )

        self.assertEqual(len(dataset), 1)
        self.assertEqual(dataset[0][0], 'testtitel')

        # without title
        resource = aq_base(resource)

        dataset = exports.reservations.dataset(
            {resource.uuid(): resource}, 'en', 'all', 'all'
        )

        self.assertEqual(len(dataset), 1)
        self.assertEqual(dataset[0][0], None)

    def test_export_view(self):

        class MyExportView(ExportView):
            pass

        view = MyExportView(self.portal, self.request())

        self.assertRaises(NotImplementedError, lambda: view.content_type)
        self.assertRaises(NotImplementedError, lambda: view.file_extension)

        view.request['source'] = 'reservations'
        self.assertTrue(hasattr(view.source, '__call__'))

        view.request['source'] = 'united-reservations'
        self.assertTrue(hasattr(view.source, '__call__'))

        view.request['source'] = 'nonexistant'
        self.assertRaises(NotImplementedError, lambda: view.source)

        class MyJsonExportView(ExportView):
            content_type = 'application/json'
            file_extension = 'json'

        view = MyJsonExportView(self.portal, self.request())

        view.request['source'] = 'reservations'
        self.assertEqual('[]', view.render())

        response = view.request.RESPONSE
        self.assertEqual(response.headers['content-length'], '2')
        self.assertEqual(
            response.headers['content-disposition'],
            'filename="Plone site.Reservations (Normal).json"'
        )
        self.assertEqual(
            response.headers['content-type'], 'application/json;charset=utf-8'
        )



import json

from calendar import Calendar
from datetime import date, timedelta, datetime

from five import grok
from zope.interface import Interface
from plone import api
from plone.memoize import view

from seantis.reservation import _
from seantis.reservation import Session
from seantis.reservation import db
from seantis.reservation import utils
from seantis.reservation.models import Allocation, Reservation
from seantis.reservation.reports import GeneralReportParametersMixin
from seantis.reservation.interfaces import ISeantisReservationSpecific

calendar = Calendar()


class MonthlyReportView(
    grok.View, GeneralReportParametersMixin
):

    permission = 'seantis.reservation.ViewReservations'

    grok.require(permission)

    grok.context(Interface)
    grok.layer(ISeantisReservationSpecific)
    grok.name('monthly_report')  # note that this text is copied in utils.py

    template = grok.PageTemplateFile('../templates/monthly_report.pt')

    @property
    def title(self):
        return _(
            u'Monthly Report for ${month} ${year}',
            mapping={
                'month': utils.month_name(
                    self.context, self.request, self.month
                ),
                'year': self.year
            }
        )

    @property
    def year(self):
        return int(self.request.get('year', date.today().year))

    @property
    def month(self):
        return int(self.request.get('month', date.today().month))

    @property
    @view.memoize
    def min_hour(self):
        return min(
            (r.first_hour for r in self.resources.values()
                if hasattr(r, 'first_hour'))
        )

    @property
    @view.memoize
    def max_hour(self):
        return max(
            (r.last_hour for r in self.resources.values()
                if hasattr(r, 'last_hour'))
        )

    @property
    @view.memoize
    def results(self):
        return monthly_report(
            self.year, self.month, self.resources, self.reservations or '*'
        )

    @property
    def show_timetable(self):
        # hide_timetable is the query parameter as showing is the default
        return False if self.request.get('hide_timetable') else True

    def build_url(self, year, month):
        params = [
            ('year', str(year)),
            ('month', str(month))
        ]

        if not self.show_timetable:
            params.append(('hide_timetable', '1'))

        return super(MonthlyReportView, self).build_url(
            extra_parameters=params
        )

    @property
    def forward_url(self):
        year, month = self.year, self.month

        if month == 12:
            year += 1
            month = 1
        else:
            month += 1

        return self.build_url(year, month)

    @property
    def backward_url(self):
        year, month = self.year, self.month

        if month == 1:
            year -= 1
            month = 12
        else:
            month -= 1

        return self.build_url(year, month)

    def format_day(self, day):
        daydate = datetime(self.year, self.month, day)
        weekday = utils.weekdayname_abbr(
            self.context, self.request, utils.shift_day(daydate.weekday())
        )

        return ', '.join((
            weekday,
            api.portal.get_localized_time(daydate, long_format=False)
        ))

    def has_reservations(self, resource):
        for status in resource['lists']:
            if resource[status]:
                return True

        return False


def monthly_report(year, month, resources, reservations='*'):

    schedulers, titles = dict(), dict()

    for uuid in resources.keys():
        schedulers[uuid] = db.Scheduler(uuid)
        titles[uuid] = utils.get_resource_title(resources[uuid])

    # this order is used for every day in the month
    ordered_uuids = [i[0] for i in sorted(titles.items(), key=lambda i: i[1])]

    # build the hierarchical structure of the report data
    report = utils.OrderedDict()
    last_day = 28

    for d in sorted((d for d in calendar.itermonthdates(year, month))):
        if not d.month == month:
            continue

        day = d.day
        last_day = max(last_day, day)
        report[day] = utils.OrderedDict()

        for uuid in ordered_uuids:
            report[day][uuid] = dict()
            report[day][uuid][u'title'] = titles[uuid]
            report[day][uuid][u'approved'] = list()
            report[day][uuid][u'pending'] = list()
            report[day][uuid][u'url'] = resources[uuid].absolute_url()
            report[day][uuid][u'lists'] = {
                u'approved': _(u'Approved'),
                u'pending': _(u'Pending'),
            }

    # gather the reservations with as much bulk loading as possible
    period_start = date(year, month, 1)
    period_end = date(year, month, last_day)

    # get a list of relevant allocations in the given period
    query = Session.query(Allocation)
    query = query.filter(period_start <= Allocation._start)
    query = query.filter(Allocation._start <= period_end)
    query = query.filter(Allocation.resource == Allocation.mirror_of)
    query = query.filter(Allocation.resource.in_(resources.keys()))

    allocations = query.all()

    # quit if there are no allocations at this point
    if not allocations:
        return {}

    # store by group as it will be needed multiple times over later
    groups = dict()
    for allocation in allocations:
        groups.setdefault(allocation.group, list()).append(allocation)

    # using the groups get the relevant reservations
    query = Session.query(Reservation)
    query = query.filter(Reservation.target.in_(groups.keys()))

    if reservations != '*':
        query = query.filter(Reservation.token.in_(reservations))

    query = query.order_by(Reservation.status)

    reservations = query.all()

    @utils.memoize
    def json_timespans(start, end):
        return json.dumps([dict(start=start, end=end)])

    used_days = dict([(i, False) for i in range(1, 32)])

    def add_reservation(start, end, reservation):
        day = start.day

        used_days[day] = True

        end += timedelta(microseconds=1)
        start = api.portal.get_localized_time(start, time_only=True)
        end = api.portal.get_localized_time(end, time_only=True)

        context = resources[utils.string_uuid(reservation.resource)]

        reservation_lists = report[day][utils.string_uuid(
            reservation.resource
        )]
        reservation_lists[reservation.status].append(
            dict(
                start=start,
                end=end,
                email=reservation.email,
                data=reservation.data,
                timespans=json_timespans(start, end),
                token=reservation.token,
                quota=utils.get_reservation_quota_statement(reservation.quota),
                resource=context
            )
        )

    for reservation in reservations:
        if reservation.target_type == u'allocation':
            add_reservation(reservation.start, reservation.end, reservation)
        else:
            for allocation in groups[reservation.target]:
                add_reservation(allocation.start, allocation.end, reservation)

    # remove unused days
    for day in report:
        if not used_days[day]:
            del report[day]

    return report



from textwrap import dedent

from datetime import date, datetime, timedelta

from seantis.reservation import utils
from seantis.reservation import settings
from seantis.reservation.error import CustomReservationError
from seantis.reservation.tests import IntegrationTestCase
from seantis.reservation.restricted_eval import (
    validate_expression,
    evaluate_expression,
    run_pre_reserve_script
)


class TestCustomFilterScript(IntegrationTestCase):

    def test_restricted_validate(self):
        run = lambda script: validate_expression(dedent(script), 'exec')

        # wrong syntax
        script = """
        if True
            pass
        """
        self.assertRaises(SyntaxError, run, script)

        # disallowed opcode
        script = """
        class Test(object):
            pass
        """
        self.assertRaises(ValueError, run, script)

    def test_restricted_evaluate(self):
        run = lambda s, g=None, l=None: evaluate_expression(
            dedent(s), g, l, 'exec'
        )

        script = "file = open('x')"
        self.assertRaises(NameError, run, script)

        locals_ = {}
        run(script, g={'open': lambda x: 'foo'}, l=locals_)
        self.assertEqual(locals_['file'], 'foo')

        run("file = var", g={'var': 'test'}, l=locals_)
        self.assertEqual(locals_['file'], 'test')

    def test_pre_reserve_script_variables(self):

        class MockContext(object):
            def getPhysicalPath(self):
                return ['', 'foo', 'bar']

        start, end = datetime.now(), datetime.now()
        data = utils.mock_data_dictionary({
            'is_test': True,
            'description': 'This is the description'
        })

        settings.set('pre_reservation_script', u'exit()')

        locals_ = {}
        run_pre_reserve_script(MockContext(), start, end, data, locals_)

        self.assertEqual(locals_['path'], '/foo/bar')
        self.assertFalse(locals_['is_group_reservation'])
        self.assertEqual(locals_['start'], start)
        self.assertEqual(locals_['end'], end)
        self.assertEqual(locals_['date'], date)
        self.assertEqual(locals_['datetime'], datetime)
        self.assertEqual(locals_['timedelta'], timedelta)
        self.assertTrue(locals_['formset_available']('mock'))

    def test_pre_reserve_script_example(self):

        class MockContext(object):
            def getPhysicalPath(self):
                return ['', 'foo', 'bar']

        script = u"""
        if is_group_reservation:
            exit()

        if not formset_available('personalien'):
            exit()

        if personalien.zipcode != '1337':
            error('People with uncool zipcodes are not welcome.')
        """

        settings.set('pre_reservation_script', dedent(script))

        start, end = None, None
        data = utils.mock_data_dictionary({})

        run = lambda: run_pre_reserve_script(MockContext(), start, end, data)

        run()  # exits early because it's a group reservation

        start, end = datetime.now(), datetime.now()

        run()  # exits early because the formset is not available

        data = utils.mock_data_dictionary({
            'zipcode': '1234'
        }, 'personalien')

        self.assertRaises(CustomReservationError, run)  # uncool zipcode

        data = utils.mock_data_dictionary({
            'zipcode': '1337'
        }, 'personalien')

        run()  # ok



import codecs
import isodate

from copy import copy
from collections import namedtuple
from datetime import datetime, date, time

from five import grok
from zope import schema
from zope.interface import Interface
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm

from z3c.form import field
from z3c.form import button

from plone import api
from plone.app.textfield.value import RichTextValue

from seantis.plonetools import tools

from seantis.reservation import _
from seantis.reservation import form
from seantis.reservation import utils
from seantis.reservation import exports
from seantis.reservation.form import extract_action_data
from seantis.reservation.base import BaseView, BaseForm

Source = namedtuple('Source', ['id', 'title', 'description', 'method'])

sources = [
    Source(
        'reservations', _(u'Reservations (Normal)'),
        _(
            u'The default reservations export with every date in the resource '
            u'having a separate record.'
        ),
        lambda resources, language, year, month, transform_record:
        exports.reservations.dataset(
            resources, language, year, month, transform_record, compact=False
        )
    ),

    Source(
        'united-reservations', _(u'Reservations (Compact)'),
        _(
            u'Like the normal reservations export, but with '
            u'group-reservations spanning multiple days merged into single '
            u'records.'
        ),
        lambda resources, language, year, month, transform_record:
        exports.reservations.dataset(
            resources, language, year, month, transform_record, compact=True
        )
    )
]


extensions = {
    'xls': _(u'Excel Format (XLS)'),
    'xlsx': _(u'Excel Format (XLSX)'),
    'csv': _(u'CSV Format'),
    'json': _(u'JSON Format'),
}


def get_sources_description(request):
    translate = tools.translator(request, 'seantis.reservation')

    items = [
        u'<dt>{}</dt><dd>{}</dd>'.format(
            *map(translate, (s.title, s.description))
        ) for s in sources
    ]

    return u'<dl>{}</dl>'.format(''.join(items))


@grok.provider(IContextSourceBinder)
def year_choices(context):
    years = [
        SimpleTerm(value=u'all', title=_(u'All'))
    ]

    minyear = datetime.now().year - 5
    maxyear = datetime.now().year + 1

    for year in range(minyear, maxyear + 1):
        years.append(
            SimpleTerm(value=unicode(year), title=unicode(year))
        )

    return SimpleVocabulary(years)


month_choices = SimpleVocabulary(
    [
        SimpleTerm(value=u'all', title=_(u'All')),
        SimpleTerm(value=u'1', title=_(u'January')),
        SimpleTerm(value=u'2', title=_(u'February')),
        SimpleTerm(value=u'3', title=_(u'March')),
        SimpleTerm(value=u'4', title=_(u'April')),
        SimpleTerm(value=u'5', title=_(u'May')),
        SimpleTerm(value=u'6', title=_(u'June')),
        SimpleTerm(value=u'7', title=_(u'July')),
        SimpleTerm(value=u'8', title=_(u'August')),
        SimpleTerm(value=u'9', title=_(u'September')),
        SimpleTerm(value=u'10', title=_(u'October')),
        SimpleTerm(value=u'11', title=_(u'November')),
        SimpleTerm(value=u'12', title=_(u'December')),
    ]
)

export_choices = SimpleVocabulary(
    [SimpleTerm(value=source.id, title=source.title) for source in sources]
)

format_choices = SimpleVocabulary(
    [
        SimpleTerm(value=ext, title=name)
        for ext, name in sorted(extensions.items(), key=lambda i: i[1])
    ]
)


class IExportSelection(Interface):
    """ Configuration for all exports. """

    export = schema.Choice(
        title=_(u'Export'),
        source=export_choices,
        required=True,
        default=sources[0].id,
    )

    format = schema.Choice(
        title=_(u'Format'),
        source=format_choices,
        required=True,
        default=extensions.keys()[0]
    )

    year = schema.Choice(
        title=_(u"Year"),
        source=year_choices,
        required=True,
        default=u'all'
    )

    month = schema.Choice(
        title=_(u"Month"),
        source=month_choices,
        required=True,
        default=u'all'
    )


class ExportSelection(BaseForm, form.ResourceParameterView):
    permission = 'seantis.reservation.ViewReservations'
    grok.require(permission)
    grok.context(Interface)
    grok.name('reservation-exports')

    label = _(u'Reservation Export')

    fields = field.Fields(IExportSelection)
    ignoreContext = True

    enable_unload_protection = False

    @property
    def action(self):
        return u'{base}/reservation-exports?uuid={uuids}'.format(
            base=self.context.absolute_url(),
            uuids='&uuid='.join(self.uuids)
        )

    def build_export_url(self, data):
        if not self.uuids:
            utils.form_error(_(u"Missing 'uuid' parameter"))

        url_template = (
            u'{base}/reservation-export.{format}?source={export}&uuid={uuids}'
            u'&year={year}&month={month}'
        )

        data['base'] = self.context.absolute_url()
        data['uuids'] = '&uuid='.join(self.uuids)

        return url_template.format(**data)

    def updateActions(self):
        super(ExportSelection, self).updateActions()
        self.actions['export'].addClass('allowMultiSubmit')
        self.actions['export'].addClass('context')

    def update(self):
        self.fields['export'].field = copy(self.fields['export'].field)
        self.fields['export'].field.description = get_sources_description(
            self.request
        )
        super(ExportSelection, self).update()

    @button.buttonAndHandler(_(u'Export'))
    @extract_action_data
    def export(self, data):
        self.request.response.redirect(self.build_export_url(data))

    @button.buttonAndHandler(_(u'Cancel'))
    @extract_action_data
    def cancel(self, data):
        self.request.response.redirect(self.context.absolute_url())


def convert_datelikes_to_isoformat(record):
    for ix, value in enumerate(record):
        if isinstance(value, datetime):
            record[ix] = isodate.datetime_isoformat(value)
        elif isinstance(value, date):
            record[ix] = isodate.date_isoformat(value)
        elif isinstance(value, time):
            record[ix] = isodate.date_isoformat(value)


def extract_text_from_richtext(record):
    transforms = api.portal.get_tool('portal_transforms')

    for ix, value in enumerate(record):
        if isinstance(value, RichTextValue):
            stream = transforms.convertTo(
                'text/plain', value.output, mimetype='text/html'
            )
            record[ix] = stream.getData().strip()


def convert_boolean_to_yes_no(record):
    for ix, value in enumerate(record):
        if isinstance(value, bool):
            record[ix] = _(u'Yes') if value is True else _(u'No')


def prepare_record(record, target_format):

    if target_format in ('xls', 'xlsx'):
        convert_boolean_to_yes_no(record)

    convert_datelikes_to_isoformat(record)
    extract_text_from_richtext(record)

    return record


class ExportView(BaseView, form.ResourceParameterView):
    """Exports the reservations from a list of resources. """

    permission = 'seantis.reservation.ViewReservations'
    grok.require(permission)

    grok.context(Interface)
    grok.baseclass()

    def get_source_by_id(self, source_id):
        source = next((s for s in sources if s.id == source_id), None)

        if not source:
            raise NotImplementedError

        return source

    @property
    def content_type(self):
        raise NotImplementedError

    @property
    def file_extension(self):
        raise NotImplementedError

    @property
    def language(self):
        return self.request.get('lang', 'en')

    @property
    def year(self):
        return self.request.get('year', 'any')

    @property
    def month(self):
        return self.request.get('month', 'any')

    @property
    def source(self):
        source = self.get_source_by_id(self.request.get('source'))
        transform_record = lambda r: prepare_record(r, self.file_extension)

        return lambda: source.method(
            self.resources,
            self.language,
            self.year,
            self.month,
            transform_record
        )

    @property
    def filename(self):
        parts = []

        parts.append(self.context.title)

        source = self.get_source_by_id(self.request.get('source'))
        translate = tools.translator(self.request, 'seantis.reservation')
        parts.append(translate(source.title))

        if self.year not in ('any', 'all'):
            parts.append(self.year)

        if self.month not in ('any', 'all'):
            if len(self.month) == 1:
                parts.append('0{}'.format(self.month))
            else:
                parts.append(self.month)

        parts.append(self.file_extension)

        return '.'.join(parts)

    def render(self, **kwargs):
        output = getattr(self.source(), self.file_extension)

        RESPONSE = self.request.RESPONSE
        RESPONSE.setHeader(
            "Content-disposition",
            codecs.utf_8_encode('filename="{}"'.format(self.filename))[0]
        )
        RESPONSE.setHeader(
            "Content-Type", "{};charset=utf-8".format(self.content_type)
        )
        RESPONSE.setHeader("Content-Length", len(output))

        return output


class XlsExportView(ExportView):
    grok.name('reservation-export.xls')
    content_type = 'application/xls'
    file_extension = 'xls'


class XlsxExportView(ExportView):
    grok.name('reservation-export.xlsx')
    content_type = 'application/xlsx'
    file_extension = 'xlsx'


class JsonExportView(ExportView):
    grok.name('reservation-export.json')
    content_type = 'application/json'
    file_extension = 'json'


class CsvExportView(ExportView):
    grok.name('reservation-export.csv')
    content_type = 'application/csv'
    file_extension = 'csv'



from sqlalchemy.ext import declarative
ORMBase = declarative.declarative_base()

from zope.i18nmessageid import MessageFactory
_ = MessageFactory('seantis.reservation')

from session import Session

# Pyflakes
Session



from five import grok

from plone.app.layout.globals.layout import LayoutPolicy
from plone.directives.form import Form

from seantis.plonetools import tools
from seantis.reservation.interfaces import ISeantisReservationSpecific


class BaseView(grok.View):

    grok.baseclass()
    grok.layer(ISeantisReservationSpecific)

    def translate(self, text):
        return tools.translator(self.request, 'seantis.reservation')(text)


class BaseViewlet(grok.Viewlet):

    grok.baseclass()
    grok.layer(ISeantisReservationSpecific)


class BaseForm(Form):

    grok.baseclass()
    grok.layer(ISeantisReservationSpecific)


class ReservationLayoutPolicy(LayoutPolicy):

    def bodyClass(self, template, view):
        """Returns the CSS class to be used on the body tag.
        """

        body_class = LayoutPolicy.bodyClass(self, template, view)

        additional_classes = [
            'seantis-reservation-view'
        ]

        if hasattr(view, 'body_classes'):
            view_body_classes = view.body_classes

            if view_body_classes:
                additional_classes.extend(view.body_classes)

        return '{} {}'.format(body_class, ' '.join(additional_classes))



from datetime import datetime, timedelta, date

from seantis.reservation import utils
from seantis.reservation import settings
from seantis.reservation.tests import IntegrationTestCase


class UtilsTestCase(IntegrationTestCase):

    def test_pairs(self):
        one = ('aa', 'bb', 'cc', 'dd')
        two = (('aa', 'bb'), ('cc', 'dd'))

        self.assertEqual(utils.pairs(one), utils.pairs(two))

    def test_event_class(self):
        self.assertEqual(utils.event_class(100), 'event-available')
        self.assertEqual(utils.event_class(75), 'event-available')
        self.assertEqual(utils.event_class(1), 'event-partly-available')
        self.assertEqual(utils.event_class(0), 'event-unavailable')

        settings.set('available_threshold', 100)
        settings.set('partly_available_threshold', 1)

        self.assertEqual(utils.event_class(100), 'event-available')
        self.assertEqual(utils.event_class(99), 'event-partly-available')
        self.assertEqual(utils.event_class(1), 'event-partly-available')
        self.assertEqual(utils.event_class(0), 'event-unavailable')

    def test_merge_additional_data(self):
        base, extra = {}, {}

        self.assertEqual(utils.merge_data_dictionaries(base, extra), {})

        base = {
            'form': {
                'desc': "Base Description",
                'interface': "Base Interface",
                'values': [
                    {
                        'key': "one",
                        'value': "eins",
                        'desc': "one == eins"
                    },
                    {
                        'key': "two",
                        'value': "zwei",
                        'desc': "two == zwei"
                    },
                ]
            }
        }

        extra = {
            'form': {
                'desc': "Extra Description",
                'interface': "Extra Interface",
                'values': [
                    {
                        'key': "two",
                        'value': "deux",
                        'desc': "two == deux"
                    },
                    {
                        'key': "three",
                        'value': "drei",
                        'desc': "three == drei"
                    },
                ]
            },
            'another_form': {}
        }

        merged = {
            'form': {
                'desc': "Extra Description",
                'interface': "Extra Interface",
                'values': [
                    {
                        'key': "one",
                        'value': "eins",
                        'desc': "one == eins"
                    },
                    {
                        'key': "two",
                        'value': "deux",
                        'desc': "two == deux"
                    },
                    {
                        'key': "three",
                        'value': "drei",
                        'desc': "three == drei"
                    },
                ]
            },
            'another_form': {}
        }

        self.maxDiff = None
        actual_merged = utils.merge_data_dictionaries(base, extra)

        self.assertEqual(merged.keys(), actual_merged.keys())
        self.assertEqual(
            merged['form']['values'], actual_merged['form']['values']
        )
        self.assertEqual(merged['form'], actual_merged['form'])

    def test_as_human_readable_string(self):
        self.assertEqual(
            utils.as_human_readable_string(datetime(2012, 1, 12)),
            '12.01.2012 00:00'
        )

        self.assertEqual(
            utils.as_human_readable_string(datetime(1889, 5, 12, 12)),
            '12.05.1889 12:00'
        )

        self.assertEqual(
            utils.as_human_readable_string(datetime(3, 5, 12, 13, 3)),
            '12.05.0003 13:03'
        )

        self.assertEqual(
            utils.as_human_readable_string(date(2012, 1, 12)),
            '12.01.2012'
        )

        self.assertEqual(
            utils.as_human_readable_string(date(1889, 5, 12)),
            '12.05.1889'
        )

        self.assertEqual(
            utils.as_human_readable_string(date(3, 5, 12)),
            '12.05.0003'
        )

        self.assertEqual(
            utils.as_human_readable_string(None), u''
        )

        self.assertEqual(
            utils.as_human_readable_string([True, False]),
            u'Yes, No'
        )

    def test_align_dates(self):
        self.assertEqual(
            utils.align_date_to_day(datetime(2012, 1, 1, 0, 0), 'down'),
            datetime(2012, 1, 1, 0, 0)
        )

        self.assertEqual(
            utils.align_date_to_day(datetime(2012, 1, 1, 0, 0), 'up'),
            datetime(2012, 1, 1, 23, 59, 59, 999999)
        )

        self.assertEqual(
            utils.align_date_to_day(datetime(2012, 1, 1, 0, 1), 'down'),
            datetime(2012, 1, 1, 0, 0)
        )

        self.assertEqual(
            utils.align_date_to_day(datetime(2012, 1, 1, 0, 1), 'up'),
            datetime(2012, 1, 1, 23, 59, 59, 999999)
        )

        self.assertEqual(
            utils.align_date_to_day(
                datetime(2012, 1, 1, 23, 59, 59, 999999), 'down'
            ),
            datetime(2012, 1, 1, 0, 0)
        )

        self.assertEqual(
            utils.align_date_to_day(
                datetime(2012, 1, 1, 23, 59, 59, 999999), 'up'
            ),
            datetime(2012, 1, 1, 23, 59, 59, 999999)
        )

        self.assertEqual(
            utils.align_range_to_day(
                datetime(2012, 1, 2, 0, 0),
                datetime(2012, 1, 2, 0, 0)
            ),
            (
                datetime(2012, 1, 2, 0, 0),
                datetime(2012, 1, 2, 23, 59, 59, 999999)
            )
        )

        self.assertEqual(
            utils.align_range_to_day(
                datetime(2012, 1, 2, 0, 0),
                datetime(2012, 1, 3, 0, 0)
            ),
            (
                datetime(2012, 1, 2, 0, 0),
                datetime(2012, 1, 3, 23, 59, 59, 999999)
            )
        )

    def test_whole_day(self):
        self.assertTrue(
            utils.whole_day(
                datetime(2012, 1, 1),
                datetime(2012, 1, 2)
            )
        )
        self.assertFalse(
            utils.whole_day(
                datetime(2012, 1, 1),
                datetime(2012, 1, 1)
            )
        )
        self.assertTrue(
            utils.whole_day(
                datetime(2012, 1, 1),
                datetime(2012, 1, 2) - timedelta(seconds=1)
            )
        )
        self.assertFalse(
            utils.whole_day(
                datetime(2012, 1, 1),
                datetime(2012, 1, 2) - timedelta(seconds=2)
            )
        )
        self.assertTrue(
            utils.whole_day(
                datetime(2012, 1, 1),
                datetime(2012, 1, 3)
            )
        )
        self.assertTrue(
            utils.whole_day(
                datetime(2012, 1, 1),
                datetime(2012, 1, 3) - timedelta(seconds=1)
            )
        )
        self.assertFalse(
            utils.whole_day(
                datetime(2012, 1, 1),
                datetime(2012, 1, 3) - timedelta(seconds=2)
            )
        )

    def test_unite(self):
        self.assertEqual(
            utils.unite(
                [1, 1, 1, 2, 2], lambda last, current: last == current
            ),
            [[1, 1, 1], [2, 2]]
        )

        united = utils.United(lambda last, current: last == current)

        united.append(1)
        self.assertEqual(united.groups, [[1]])

        united.append(1)
        self.assertEqual(united.groups, [[1, 1]])

        united.append(2)
        self.assertEqual(united.groups, [[1, 1], [2]])

        united.append(1)
        self.assertEqual(united.groups, [[1, 1], [2], [1]])

    def test_unite_dates(self):
        self.assertEqual(
            list(utils.unite_dates([
                (datetime(2012, 1, 1), datetime(2012, 1, 2))
            ])),
            [
                (datetime(2012, 1, 1), datetime(2012, 1, 2))
            ]
        )
        self.assertEqual(
            list(utils.unite_dates([
                (datetime(2012, 1, 1), datetime(2012, 1, 2)),
                (datetime(2012, 1, 3), datetime(2012, 1, 4))
            ])),
            [
                (datetime(2012, 1, 1), datetime(2012, 1, 2)),
                (datetime(2012, 1, 3), datetime(2012, 1, 4))
            ]
        )
        self.assertEqual(
            list(utils.unite_dates([
                (datetime(2012, 1, 1), datetime(2012, 1, 2)),
                (datetime(2012, 1, 2), datetime(2012, 1, 3)),
                (datetime(2012, 1, 4), datetime(2012, 1, 5)),
            ])),
            [
                (datetime(2012, 1, 1), datetime(2012, 1, 3)),
                (datetime(2012, 1, 4), datetime(2012, 1, 5)),
            ]
        )
        self.assertEqual(
            list(utils.unite_dates([
                (datetime(2012, 1, 1), datetime(2012, 1, 2)),
                (datetime(2012, 1, 3), datetime(2012, 1, 4)),
                (datetime(2012, 1, 4), datetime(2012, 1, 5)),
            ])),
            [
                (datetime(2012, 1, 1), datetime(2012, 1, 2)),
                (datetime(2012, 1, 3), datetime(2012, 1, 5)),
            ]
        )
        self.assertEqual(
            list(utils.unite_dates([
                (datetime(2012, 1, 1), datetime(2012, 1, 2)),
                (datetime(2012, 1, 2), datetime(2012, 1, 3)),
                (datetime(2012, 1, 10), datetime(2012, 1, 11)),
                (datetime(2012, 1, 3), datetime(2012, 1, 4)),
                (datetime(2012, 1, 11), datetime(2012, 1, 12)),
            ])),
            [
                (datetime(2012, 1, 1), datetime(2012, 1, 4)),
                (datetime(2012, 1, 10), datetime(2012, 1, 12)),
            ]
        )



# -*- coding: utf-8 -*-

from logging import getLogger
log = getLogger('seantis.reservation')

from copy import copy
from datetime import time
from DateTime import DateTime
from five import grok

from plone.dexterity.interfaces import IDexterityFTI
from zope.component import queryUtility
from zope.interface import Interface
from zope.security import checkPermission

from z3c.form import field
from z3c.form import button
from z3c.form.browser.radio import RadioFieldWidget
from z3c.form.browser.checkbox import CheckBoxFieldWidget
from zExceptions import NotFound
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.schema import Choice, List, Set

from seantis.reservation.throttle import throttled
from seantis.reservation.interfaces import (
    IResourceBase,
    IReservation,
    IGroupReservation,
    IRevokeReservation,
    IReservationIdForm,
    ISeantisReservationSpecific
)

from seantis.reservation.base import BaseView, BaseViewlet
from seantis.reservation.error import DirtyReadOnlySession
from seantis.reservation.restricted_eval import run_pre_reserve_script
from seantis.reservation import _
from seantis.reservation import db
from seantis.reservation import utils
from seantis.reservation import plone_session
from seantis.reservation.form import (
    ResourceBaseForm,
    AllocationGroupView,
    ReservationListView,
    extract_action_data
)

from seantis.reservation.overview import OverviewletManager
from seantis.reservation.error import NoResultFound


class ReservationUrls(object):
    """ Mixin class to create admin URLs for a specific reservation. """

    def revoke_all_url(self, token, context=None):
        context = context or self.context
        base = context.absolute_url()
        return base + u'/revoke-reservation?reservation=%s' % token

    def approve_all_url(self, token, context=None):
        context = context or self.context
        base = context.absolute_url()
        return base + u'/approve-reservation?reservation=%s' % token

    def deny_all_url(self, token, context=None):
        context = context or self.context
        base = context.absolute_url()
        return base + u'/deny-reservation?reservation=%s' % token

    def update_all_url(self, token, context=None):
        context = context or self.context
        base = context.absolute_url()
        return base + u'/update-reservation-data?reservation=%s' % token

    def print_all_url(self, token, context):
        context = context or self.context
        base = context.absolute_url()
        return base + u'/reservations?reservation={}&print=1'.format(token)

    def show_all_url(self, token, context):
        context = context or self.contex
        base = context.absolute_url()
        return base + u'/reservations?reservation={}'.format(token)


class ReservationSchemata(object):
    """ Mixin to use with plone.autoform and IResourceBase which makes the
    form it is used on display the formsets defined by the user.

    A formset is a Dexterity Type defined through the admin interface or
    code which has the behavior IReservationFormset.

    """

    @property
    def may_view_manager_sets(self):
        manager_permission = 'seantis.reservation.EditReservations'
        return checkPermission(manager_permission, self.context)

    def is_manager_set(self, fti):
        behavior = 'seantis.reservation.interfaces.IReservationManagerFormSet'
        return behavior in fti.behaviors

    @property
    def additionalSchemata(self):
        scs = []
        self.fti = dict()

        for ptype in self.context.formsets:
            fti = queryUtility(IDexterityFTI, name=ptype)
            if fti:
                if self.is_manager_set(fti) and not self.may_view_manager_sets:
                    continue  # do not show, but fill with defaults later

                schema = fti.lookupSchema()
                scs.append((ptype, fti.title, schema))

                self.fti[ptype] = (fti.title, schema)

        return scs


class SessionFormdataMixin(ReservationSchemata):

    def email(self, form_data=None):

        if not form_data or not form_data.get('email'):
            email = plone_session.get_email(self.context)
        else:
            email = form_data['email']
            plone_session.set_email(self.context, email)

        return email

    def merge_formdata(self, existing, new):

        for form in new:
            existing[form] = new[form]

        return existing

    @property
    def manager_ftis(self):
        ftis = {}

        for ptype in self.context.formsets:
            fti = queryUtility(IDexterityFTI, name=ptype)
            if fti and self.is_manager_set(fti):
                ftis[ptype] = (fti.title, fti.lookupSchema())

        return ftis

    def additional_data(self, form_data=None, add_manager_defaults=False):

        if not form_data:
            data = plone_session.get_additional_data(self.context)
        else:
            data = plone_session.get_additional_data(self.context) or dict()

            # merge the formdata for session use only, committing the
            # reservation only forms defined in the resource are
            # stored with the reservation to get proper separation
            data = self.merge_formdata(
                plone_session.get_additional_data(self.context) or dict(),
                utils.additional_data_dictionary(form_data, self.fti)
            )

            plone_session.set_additional_data(self.context, data)

        # the default values of manager forms are added to users without
        # the permission right before saving
        if add_manager_defaults and not self.may_view_manager_sets:
            defaults = {}
            manager_ftis = self.manager_ftis

            for key, info in manager_ftis.items():
                for name, f in field.Fields(info[1]).items():
                    if f.field.default is not None:
                        fieldkey = '{}.{}'.format(key, name)
                        defaults[fieldkey] = f.field.default

            data = self.merge_formdata(
                data, utils.additional_data_dictionary(defaults, manager_ftis)
            )

        # on the other hand, if the user is not allowed, the data is cleared,
        # just in case (really more of a dev-environment problem, but it
        # doesn't hurt anyway)
        if data:
            if not add_manager_defaults and not self.may_view_manager_sets:
                manager_ftis = self.manager_ftis

                for form in data.keys():
                    if form in manager_ftis:
                        del data[form]

        return data

    def session_id(self):
        return plone_session.get_session_id(self.context)


class YourReservationsData(object):
    """ Mixin providing functions to deal with 'your' reservations. """

    def reservations(self):
        """ Returns all reservations in the user's session """
        session_id = plone_session.get_session_id(self.context)
        return db.reservations_by_session(session_id).all()

    def resources(self):
        """ Returns a list of resources contained in the reservations. The
        result is a set of uuid strings (without hyphens).

        """
        return set(utils.string_uuid(r.resource) for r in self.reservations())

    @property
    def has_reservations(self):
        session_id = plone_session.get_session_id(self.context)
        return bool(db.reservations_by_session(session_id).first())

    def confirm_reservations(self, token=None):
        # Remove session_id from all reservations in the current session.
        db.confirm_reservations_for_session(
            plone_session.get_session_id(self.context),
            token,
            utils.get_current_language(self.context, self.request)
        )

    def remove_reservation(self, token):
        try:
            session_id = plone_session.get_session_id(self.context)
            db.remove_reservation_from_session(session_id, token)
        except NoResultFound:
            pass  # act idempotent to the user

    def reservation_data(self):
        """ Prepares data to be shown in the my reservation's table """
        reservations = []

        for reservation in self.reservations():
            resource = utils.get_resource_by_uuid(reservation.resource)

            if resource is None:
                log.warn('Invalid UUID %s' % str(reservation.resource))
                continue

            resource = resource.getObject()

            data = {}

            data['title'] = utils.get_resource_title(resource)

            timespans = []
            for start, end in reservation.timespans():
                timespans.append(utils.display_date(start, end))

            data['time'] = '<ul class="dense"><li>{}</li></ul>'.format(
                '</li><li>'.join(timespans)
            )
            data['quota'] = utils.get_reservation_quota_statement(
                reservation.quota
            ) if reservation.quota > 1 else u''

            data['url'] = resource.absolute_url()
            data['remove-url'] = ''.join((
                resource.absolute_url(),
                '/your-reservations?remove=',
                reservation.token.hex
            ))
            reservations.append(data)

        return reservations

    def redirect_to_your_reservations(self):
        self.request.response.redirect(
            self.context.absolute_url() + '/your-reservations'
        )


class ReservationBaseForm(ResourceBaseForm):

    def your_reservation_defaults(self, defaults):
        """ Extends the given dictionary containing field defaults with
        the defaults found in your-reservations.

        """

        default_email = self.email()
        if default_email:
            defaults['email'] = self.email()

        data = self.additional_data()

        if not data:
            return defaults

        for form in data:
            if form in self.context.formsets:
                for field in data[form]['values']:
                    defaults["%s.%s" % (form, field['key'])] = field['value']

        return defaults

    def run_reserve(
        self, data, approve_manually, start=None, end=None, group=None, quota=1
    ):

        assert (start and end) or group
        assert not (start and end and group)

        email = self.email(data)
        additional_data = self.additional_data(data, add_manager_defaults=True)
        session_id = self.session_id()

        # only store forms defined in the formsets list
        additional_data = dict(
            (
                form, additional_data[form]
            ) for form in self.context.formsets if form in additional_data
        )

        run_pre_reserve_script(self.context, start, end, additional_data)

        def run():
            if start and end:
                return self.scheduler.reserve(
                    email, (start, end),
                    data=additional_data, session_id=session_id, quota=quota
                )
            else:
                return self.scheduler.reserve(
                    email, group=group,
                    data=additional_data, session_id=session_id, quota=quota
                )

        token = throttled(run, 'reserve')()

        if not approve_manually:
            self.scheduler.approve_reservation(token)

        self.flash(
            _(
                u'Added reservation to your list. '
                u'You have 15 minutes to confirm your reservations.'
            )
        )


class ReservationForm(
        ReservationBaseForm,
        SessionFormdataMixin,
        YourReservationsData
):

    permission = 'seantis.reservation.SubmitReservation'

    grok.name('reserve')
    grok.require(permission)

    context_buttons = ('reserve', )
    standalone_buttons = ('cancel', )

    fields = field.Fields(IReservation)
    label = _(u'Resource reservation')

    fti = None

    autoGroups = True
    enable_form_tabbing = True
    default_fieldset_label = _(u'General Information')

    @property
    def css_class(self):
        return super(ReservationForm, self).css_class + ' next-next-wizard'

    @property
    def hidden_fields(self):
        hidden = ['id']

        try:
            allocation = self.allocation(self.id)

            if allocation:

                if allocation.reservation_quota_limit == 1:
                    hidden.append('quota')

                if allocation.whole_day:
                    if not allocation.partly_available:
                        hidden.append('start_time')
                        hidden.append('end_time')

        except DirtyReadOnlySession:
            pass

        except NotFound:
            pass

        return hidden

    @property
    def disabled_fields(self):
        disabled = ['day']
        try:
            allocation = self.allocation(self.id)

            if allocation:

                if not allocation.partly_available:
                    disabled.append('start_time')
                    disabled.append('end_time')

        except DirtyReadOnlySession:
            pass

        except NotFound:
            pass

        return disabled

    def defaults(self, **kwargs):
        defaults = self.your_reservation_defaults(dict(id=self.id))
        self.inject_missing_data(defaults)

        return defaults

    def allocation(self, id):
        if not id:
            raise NotFound

        try:
            return self.scheduler.allocation_by_id(id)
        except NoResultFound:
            raise NotFound

    def strptime(self, value):
        if not value:
            return None

        if not isinstance(value, basestring):
            return value

        dt = DateTime(value)
        return time(dt.hour(), dt.minute())

    def validate(self, data):
        try:
            start, end = utils.get_date_range(
                data['day'], data['start_time'], data['end_time']
            )
            if not self.allocation(data['id']).contains(start, end):
                utils.form_error(_(u'Reservation out of bounds'))

            return start, end
        except (NoResultFound, TypeError):
            utils.form_error(_(u'Invalid reservation request'))

    def inject_missing_data(self, data, allocation=None):
        """ Adds the date and start-/end-time to the data if they are missing,
        which happens because those fields may be disabled and therefore are
        not transferred when submitting the form.

        The fields are injected into the passed dictionary, which may be
        the reservations defaults or the submitted form data.

        """
        extracted, errors = self.extractData(setErrors=False)

        # the extracted fields may contain field values which need to be
        # injected so the defaults are filled - otherwise no value is updated
        # on the disabled field
        for field in ('day', 'start_time', 'end_time'):
            if extracted.get(field) is not None:
                data[field] = extracted[field]

        # if the extracted data was not of any help the id of the allocation
        # is our last resort.
        try:
            allocation = allocation or self.allocation(data['id'])
        except DirtyReadOnlySession:
            return

        if data.get('day') is None:
            data['day'] = allocation.display_start.date()

        if data.get('start_time') is None:
            data['start_time'] = allocation.display_start.time()

        if data.get('end_time') is None:
            data['end_time'] = allocation.display_end.time()

    @button.buttonAndHandler(_(u'Reserve'))
    @extract_action_data
    def reserve(self, data):

        allocation = self.allocation(data['id'])
        approve_manually = allocation.approve_manually

        self.inject_missing_data(data, allocation)

        start, end = self.validate(data)
        quota = int(data.get('quota', 1))

        # whole day allocations don't show the start / end time which is to
        # say the data arrives with 00:00 - 00:00. we align that to the day
        if allocation.whole_day:

            if not allocation.partly_available:
                assert start == end

            if start == end:
                start, end = utils.align_range_to_day(start, end)

        def reserve():
            self.run_reserve(
                data=data, approve_manually=approve_manually,
                start=start, end=end, quota=quota
            )

        utils.handle_action(
            action=reserve, success=self.redirect_to_your_reservations
        )

    @button.buttonAndHandler(_(u'Cancel'))
    def cancel(self, action):
        self.redirect_to_context()

    def customize_fields(self, fields):
        """ This function is called by ResourceBaseForm every time fields are
        created from the schema by z3c. This allows for changes before the
        fields are properly integrated into the form.

        Here, we want to make sure that all formset schemas have sane widgets.

        """

        for field in fields.values():

            field_type = type(field.field)

            if field_type is List or field_type is Set:
                field.widgetFactory = CheckBoxFieldWidget

            elif field_type is Choice:
                field.widgetFactory = RadioFieldWidget


class GroupReservationForm(
        ReservationForm,
        AllocationGroupView,
        SessionFormdataMixin,
        YourReservationsData
):
    permission = 'seantis.reservation.SubmitReservation'

    grok.name('reserve-group')
    grok.require(permission)

    context_buttons = ('reserve', )
    standalone_buttons = ('cancel', )

    fields = field.Fields(IGroupReservation)
    label = _(u'Recurrance reservation')

    template = ViewPageTemplateFile('templates/reserve_group.pt')

    ignore_requirements = True

    autoGroups = True
    enable_form_tabbing = True
    default_fieldset_label = _(u'General Information')

    @property
    def hidden_fields(self):
        hidden = ['group']

        try:
            allocation = self.group and self.scheduler.allocations_by_group(
                self.group
            ).first()

            if allocation.reservation_quota_limit == 1:
                hidden.append('quota')

        except DirtyReadOnlySession:
            pass

        return hidden

    def defaults(self, **kwargs):
        return self.your_reservation_defaults(dict(group=self.group, quota=1))

    @button.buttonAndHandler(_(u'Reserve'))
    @extract_action_data
    def reserve(self, data):

        approve_manually = self.scheduler.allocations_by_group(data['group']) \
            .first().approve_manually

        def reserve():
            self.run_reserve(
                data=data, approve_manually=approve_manually,
                group=data['group'], quota=data['quota']
            )

        utils.handle_action(
            action=reserve, success=self.redirect_to_your_reservations
        )

    @button.buttonAndHandler(_(u'Cancel'))
    def cancel(self, action):
        self.redirect_to_context()


class YourReservations(ResourceBaseForm, YourReservationsData):

    permission = "seantis.reservation.SubmitReservation"

    grok.name('your-reservations')
    grok.require(permission)

    context_buttons = ('finish', )
    standalone_buttons = ('cancel', )

    grok.context(Interface)

    css_class = 'seantis-reservation-form'

    template = grok.PageTemplateFile('templates/your_reservations.pt')

    @button.buttonAndHandler(_(u'Submit Reservations'), name="finish")
    def finish(self, data):

        resources = self.resources()

        def on_success():
            self.request.response.redirect(
                '{context}/thank-you-for-reserving?uuid={uuids}'.format(
                    context=self.context.absolute_url(),
                    uuids='&uuid='.join(resources)
                )
            )

        utils.handle_action(self.confirm_reservations, success=on_success)

    # the button's name is 'cancel' because it should behave like a cancel
    # button in the browser (namely, it should not refetch the events)
    @button.buttonAndHandler(_(u'Reserve More'), name="cancel")
    def proceed(self, data):
        # Don't do anything, reservations stay in the session.
        self.request.response.redirect(self.context.absolute_url())

    def update(self):
        if 'remove' in self.request and utils.is_uuid(self.request['remove']):
            self.remove_reservation(self.request['remove'])

            self.request.response.redirect(self.context.absolute_url())

        super(YourReservations, self).update()


class YourReservationsViewlet(BaseViewlet, YourReservationsData):
    grok.context(Interface)
    grok.name('seantis.reservation.YourReservationsviewlet')
    grok.require('zope2.View')
    grok.viewletmanager(OverviewletManager)

    grok.order(0)

    template = grok.PageTemplateFile('templates/your_reservations_viewlet.pt')

    def available(self):
        return self.has_reservations

    def finish_url(self):
        return self.context.absolute_url() + '/your-reservations'


class ReservationIdForm(ResourceBaseForm):
    """ Describes a form with a hidden reservation field and the ability to
    set the reservation using a query parameter:

    example-form?reservation=298c6de470f94c64928c14246f3ee9e5

    """

    grok.baseclass()
    fields = field.Fields(IReservationIdForm)
    hidden_fields = ('reservation', )
    extracted_data = {}

    @property
    def reservation(self):
        return self.request.get(
            'reservation', self.extracted_data.get('reservation')
        )

    def defaults(self):
        return dict(reservation=self.reservation)


class ReservationDecisionForm(ReservationIdForm, ReservationListView,
                              ReservationUrls):
    """ Base class for admin's approval / denial forms. """

    grok.baseclass()
    grok.layer(ISeantisReservationSpecific)

    template = ViewPageTemplateFile('templates/decide_reservation.pt')

    show_links = False


class ReservationApprovalForm(ReservationDecisionForm):

    permission = 'seantis.reservation.ApproveReservations'

    grok.name('approve-reservation')
    grok.require(permission)

    context_buttons = ('approve', )
    standalone_buttons = ('cancel', )

    label = _(u'Approve reservation')

    @property
    def hint(self):
        if not self.pending_reservations():
            return _(u'No such reservation')

        return _(u'Do you really want to approve the following reservations?')

    @button.buttonAndHandler(_(u'Approve'))
    @extract_action_data
    def approve(self, data):

        def approve():
            self.scheduler.approve_reservation(data['reservation'])
            self.flash(_(u'Reservation approved'))

        utils.handle_action(action=approve, success=self.redirect_to_context)

    @button.buttonAndHandler(_(u'Cancel'))
    def cancel(self, action):
        self.redirect_to_context()


class ReservationDenialForm(ReservationDecisionForm):

    permission = 'seantis.reservation.ApproveReservations'

    grok.name('deny-reservation')
    grok.require(permission)

    destructive_buttons = ('deny', )
    standalone_buttons = ('cancel', )

    label = _(u'Deny reservation')

    @property
    def hint(self):
        if not self.pending_reservations():
            return _(u'No such reservation')

        return _(u'Do you really want to deny the following reservations?')

    @button.buttonAndHandler(_(u'Deny'))
    @extract_action_data
    def deny(self, data):

        def deny():
            self.scheduler.deny_reservation(data['reservation'])
            self.flash(_(u'Reservation denied'))

        utils.handle_action(action=deny, success=self.redirect_to_context)

    @button.buttonAndHandler(_(u'Cancel'))
    def cancel(self, action):
        self.redirect_to_context()


class ReservationRevocationForm(
    ReservationIdForm,
    ReservationListView,
    ReservationUrls
):

    permission = 'seantis.reservation.ApproveReservations'

    grok.name('revoke-reservation')
    grok.require(permission)
    grok.layer(ISeantisReservationSpecific)

    destructive_buttons = ('revoke', )
    standalone_buttons = ('cancel', )

    fields = field.Fields(IRevokeReservation)
    template = ViewPageTemplateFile('templates/revoke_reservation.pt')

    label = _(u'Revoke reservation')

    show_links = False

    @property
    def hint(self):
        if not (self.reservation or self.approved_reservations()):
            return _(u'No such reservation')

        return _(
            u'Do you really want to revoke the following reservations?'
        )

    @button.buttonAndHandler(_(u'Revoke'))
    @extract_action_data
    def revoke(self, data):

        def revoke():
            self.scheduler.revoke_reservation(
                data['reservation'], data['reason'], data['send_email']
            )
            self.flash(_(u'Reservation revoked'))

        utils.handle_action(action=revoke, success=self.redirect_to_context)

    @button.buttonAndHandler(_(u'Cancel'))
    def cancel(self, action):
        self.redirect_to_context()


class ReservationList(BaseView, ReservationListView, ReservationUrls):

    permission = "seantis.reservation.ViewReservations"

    grok.name('reservations')
    grok.require(permission)

    grok.context(IResourceBase)

    template = grok.PageTemplateFile('templates/reservations.pt')

    @property
    def id(self):
        return utils.request_id_as_int(self.request.get('id'))

    @property
    def group(self):
        if 'group' in self.request:
            return unicode(self.request['group'].decode('utf-8'))
        else:
            return u''

    @property
    def reservation(self):
        """ Limits the list to the given reservation. """
        return self.request.get('reservation', None)

    @property
    def print_site(self):
        """ Returns true if the document should be printed when opening it. """
        return self.request.get('print', None) is not None

    @property
    def body_classes(self):
        if utils.is_uuid(self.reservation):
            return ['single-reservation-view']


class ReservationDataEditForm(ReservationIdForm, ReservationSchemata):

    permission = "seantis.reservation.EditReservations"

    grok.name('update-reservation-data')
    grok.require(permission)

    grok.context(IResourceBase)
    grok.layer(ISeantisReservationSpecific)

    context_buttons = ('save', )
    standalone_buttons = ('cancel', )
    extracted_errors = []

    template = ViewPageTemplateFile('templates/form.pt')

    @property
    def broken_data(self):
        return self.broken

    @property
    def label(self):
        if self.context.formsets:
            return _(u'Edit Formdata')
        else:
            return _(u'No Formdata to edit')

    def separate_broken_data(self):
        """ Goes through the reservation data and returns the the reservation
        data in two pieces. The first is the working data, the second is the
        broken data.

        Broken data is the data which cannot be loaded in the form, because
        the underlying form schema has changed since the data has been
        written.

        """
        reservation_data = self.get_reservation_data()

        def init_form(data, form):
            if form in data:
                return data

            data[form] = copy(reservation_data[form])
            data[form]['values'] = []

            return data

        working, broken = {}, {}

        for form in reservation_data:
            for value in reservation_data[form]['values']:
                id = '.'.join((form, value['key']))

                if self.get_field(id) and self.get_field(id):
                    init_form(working, form)[form]['values'].append(value)
                else:
                    init_form(broken, form)[form]['values'].append(value)

        return working, broken

    def get_reservation_data(self):
        if not self.reservation:
            return {}

        if not hasattr(self, 'reservation_data'):
            try:
                query = self.scheduler.reservation_by_token(self.reservation)
                self.reservation_data = query.one().data
            except DirtyReadOnlySession:
                self.reservation_data = {}

        return self.reservation_data

    def defaults(self):
        defaults = super(ReservationDataEditForm, self).defaults()

        if not self.context.formsets:
            return defaults

        self.working, self.broken = self.separate_broken_data()

        data = self.working

        errors = [e.widget.__name__ for e in self.extracted_errors]

        for form in data:

            for value in data[form]['values']:
                if isinstance(value['value'], basestring):
                    decoded = utils.userformdata_decode(value['value'])
                    fieldvalue = decoded or value['value']
                else:
                    fieldvalue = value['value']

                fieldkey = '{}.{}'.format(form, value['key'])
                if fieldkey in self.extracted_data or fieldkey in errors:
                    continue
                else:
                    defaults[fieldkey] = fieldvalue

        return defaults

    def customize_fields(self, fields):

        for field in fields.values():

            field_type = type(field.field)

            if field_type is List or field_type is Set:
                field.widgetFactory = CheckBoxFieldWidget

            elif field_type is Choice:
                field.widgetFactory = RadioFieldWidget

    @button.buttonAndHandler(_(u'Save'))
    @extract_action_data
    def save(self, data):

        broken = self.separate_broken_data()[1]
        working = utils.additional_data_dictionary(data, self.fti)

        if broken:
            self.additional_data = utils.merge_data_dictionaries(
                broken, working
            )
        else:
            self.additional_data = working

        def save():
            self.scheduler.update_reservation_data(
                self.reservation, self.additional_data
            )
            self.flash(_(u'Formdata updated'))

        utils.handle_action(
            action=save, success=self.redirect_to_context
        )

    @button.buttonAndHandler(_(u'Cancel'))
    def cancel(self, action):
        self.redirect_to_context()



# -*- coding: utf-8 -*-
from dateutil import rrule

from five import grok

from zope import schema
from zope.interface import Interface, invariant, Invalid, Attribute
from zope.component import getAllUtilitiesRegisteredFor as getallutils
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm

from plone.app.z3cform.wysiwyg import WysiwygFieldWidget
from plone.directives import form
from plone.dexterity.interfaces import IDexterityFTI
from plone.dexterity.utils import schemaNameToPortalType as getname

from z3c.form.browser.checkbox import CheckBoxFieldWidget
from z3c.form.browser.radio import RadioFieldWidget
from z3c.form import widget

from seantis.plonetools.schemafields import Email

from seantis.reservation import _, utils
from seantis.reservation.raster import VALID_RASTER_VALUES
from seantis.reservation.mail_templates import templates
from seantis.reservation.utils import _languagelist


class ISeantisReservationSpecific(Interface):
    pass


days = SimpleVocabulary(
    [
        SimpleTerm(value=rrule.MO, title=_(u'Mo')),
        SimpleTerm(value=rrule.TU, title=_(u'Tu')),
        SimpleTerm(value=rrule.WE, title=_(u'We')),
        SimpleTerm(value=rrule.TH, title=_(u'Th')),
        SimpleTerm(value=rrule.FR, title=_(u'Fr')),
        SimpleTerm(value=rrule.SA, title=_(u'Sa')),
        SimpleTerm(value=rrule.SU, title=_(u'Su')),
    ]
)

recurrence = SimpleVocabulary(
    [
        SimpleTerm(value=False, title=_(u'Once')),
        SimpleTerm(value=True, title=_(u'Daily')),
    ]
)

calendar_views = SimpleVocabulary(
    [
        SimpleTerm(value='month', title=_(u'Monthly View')),
        SimpleTerm(value='agendaWeek', title=_(u'Weekly View')),
        SimpleTerm(value='agendaDay', title=_(u'Daily View'))
    ]
)

default_views = ['month', 'agendaWeek', 'agendaDay']
default_selected_view = 'agendaWeek'

calendar_dates = SimpleVocabulary(
    [
        SimpleTerm(value='current', title=_(u'Always show the current date')),
        SimpleTerm(value='specific', title=_(u'Always show a specific date'))
    ]
)


def select_at_least_one(values):
    if not values:
        raise Invalid(
            _(u'Select at least one value')
        )
    return True


@grok.provider(IContextSourceBinder)
def form_interfaces(context):
    """ Used as a source for a vocabulary this function returns a vocabulary
    of interfaces which may be used as sub-forms in a resource object.

    """
    dutils = getallutils(IDexterityFTI)
    behaviors = set((
        'seantis.reservation.interfaces.IReservationFormSet',
        'seantis.reservation.interfaces.IReservationManagerFormSet'
    ))
    interfaces = [
        (
            u.title, u.lookupSchema()
        ) for u in dutils if behaviors & set(u.behaviors)
    ]

    def get_term(item):
        return SimpleTerm(title=item[0], value=getname(item[1].__name__))

    return SimpleVocabulary(map(get_term, interfaces))


@grok.provider(IContextSourceBinder)
def plone_languages(context):
    def get_term(item):
        return SimpleTerm(title=item[1]['native'], value=item[0])

    terms = sorted(map(get_term, _languagelist.items()), key=lambda t: t.title)

    return SimpleVocabulary(terms)


class IOverview(Interface):
    """ Views implementing this interface may use the OverviewletManager to
    display an overview of a list of resources.

    The OverviewletManager displays viewlets which work with a list of
    resources in a folderish view. Those resources are not required to actually
    be stored in that folder. They just need to be defined using this
    interface.

    The result may be something like this, with the part on the right side
    being the work of the overview:

    Resources Overivew

                                ╔═══════════════╗
        » Resource One          ║ ░ ░ ░ ░ ░ ░ ░ ║
        » Resource Two          ║   calendar    ║
                                ║ ░ ░ ░ ░ ░ ░ ░ ║
                                ╚═══════════════╝

                                » Monthly Report
                                » Compare Resources
    """

    def resource_map(self):
        """ Returns a dictionary mapping items to resources. Or in the case
        of a simple resource list, the list of uuids of those resources.

        Simple
        ------
        If you deal with a folderish view which only displays a list of
        resources you can simply return a list of uuids (or any other
        iterable except strings):

        [uuid1, uuid2, uuid3]

        The uuids may be UUID objects or strings.

        -> See Listing view in seantis.reservation.resource

        Grouped
        -------
        If you deal with a folderish view which shows parents of groups of
        resources you should return something like this:

        group1 -> [uuid1, uuid2]
        group2 -> [uuid3]

        This will highlight group1 if a day containing uuid1 or uuid2 is
        hovered over in the calendar. It will also highlight the days for which
        there's an allocation in any of the resources when a group is hovered
        over.

        -> See Directory View in seantis.dir.facility.directory

        Note
        ----
        For the highlighting to work the id of the element in the browser
        must be the uuid in the simple and the group key in the grouped
        example.

        """


# not really where I want this code to be, but the code needs some reorganizing
# because of circular imports and the like
class OverviewletManager(grok.ViewletManager):
    """ Manages the viewlets shown in the overview. """
    grok.context(Interface)
    grok.name('seantis.reservation.overviewletmanager')
    grok.layer(ISeantisReservationSpecific)

    @utils.cached_property
    def uuidmap(self):
        """ Returns a dictionary mapping resource uuids to item ids. """

        if not IOverview.providedBy(self.view):
            return {}

        rmap = self.view.resource_map()
        assert not isinstance(rmap, basestring)

        def transform_uuid(target):
            if utils.is_uuid(target):
                return utils.string_uuid(target)
            else:
                return target

        if not isinstance(rmap, dict):
            return dict(([transform_uuid(uuid)] * 2) for uuid in rmap)

        # overlay.js needs the map the other way around, which is mostly a
        # historical artifact, but it also makes more sense for the developer
        # using IOverview to define a group -> resources relationship
        uuidmap = {}
        for key, resources in rmap.items():
            assert isinstance(resources, (list, tuple, set))

            for resource in resources:
                uuidmap.setdefault(transform_uuid(resource), []).append(key)

        return uuidmap

    @property
    def resource_uuids(self):
        if not self.uuidmap:
            return []

        return self.uuidmap.keys()


class IReservationFormSet(Interface):
    """ Marks interface as usable for sub-forms in a resource object. """


class IReservationManagerFormSet(IReservationFormSet):
    """ Same as IReservationFormSet but only available to managers. """


class IResourceAllocationDefaults(form.Schema):

    quota = schema.Int(
        title=_(u'Quota'),
        description=_(
            u'Number of times an allocation may be reserved at the same time. '
            u'e.g. 3 spots in a daycare center, 2 available cars, '
            u'1 meeting room. '
        ),
        default=1
    )

    reservation_quota_limit = schema.Int(
        title=_(u'Reservation Quota Limit'),
        description=_(
            u'The maximum quota a single reservation may occupy at once. '
            u'There is no limit if set to zero.'
        ),
        default=1
    )

    approve_manually = schema.Bool(
        title=_(u'Manually approve reservation requests'),
        description=_(
            u'If checked a reservation manager must decide if a reservation '
            u'can be approved. Until then users are added to a waitinglist. '
            u'Reservations are automatically approved if this is not checked. '
        ),
        default=False
    )

    partly_available = schema.Bool(
        title=_(u'Partly available'),
        description=_(
            u'If the allocation is partly available users may reserve '
            u'only a part of it (e.g. half of it). If not the allocation '
            u'Must be reserved as a whole or not at all'
        ),
        default=False
    )

    raster = schema.Choice(
        title=_(u'Raster'),
        description=_(
            u'Defines the minimum length of any given reservation as well '
            u'as the alignment of the start / end of the allocation. E.g. a '
            u'raster of 30 minutes means that the allocation can only start '
            u'at xx:00 and xx:30 respectively'
        ),
        values=VALID_RASTER_VALUES,
        default=15
    )

    @invariant
    def isValidQuota(Allocation):
        if not (1 <= Allocation.quota and Allocation.quota <= 1000):
            raise Invalid(_(u'Quota must be between 1 and 1000'))

    @invariant
    def isValidQuotaLimit(Allocation):
        if Allocation.reservation_quota_limit < 0:
            raise Invalid(
                _(u'Reservation quota limit must zero or a positive number')
            )

    ######### deprecated #########
    approve = schema.Bool(
        title=_(u'DEPRECATED: Approve reservation requests'), default=True
    )
    # approve has been moved to approve_manually and will be removed in
    # a future release. approve_manually is equivalent.


class IResourceBase(IResourceAllocationDefaults):
    """ A resource displaying a calendar. """

    title = schema.TextLine(
        title=_(u'Name')
    )

    description = schema.Text(
        title=_(u'Description'),
        required=False
    )

    first_hour = schema.Int(
        title=_(u'First hour of the day'),
        description=_(
            u'Everything before this hour is not shown in the '
            u'calendar, making the calendar display more compact. '
            u'Should be set to an hour before which there cannot '
            u'be any reservations.'
        ),
        default=7
    )

    last_hour = schema.Int(
        title=_(u'Last hour of the day'),
        description=_(
            u'Everything after this hour is not shown in the '
            u'calendar, making the calendar display more compact. '
            u'Should be set to an hour after which there cannot '
            u'be any reservations.'
        ),
        default=23
    )

    available_views = schema.List(
        title=_(u'Available Views'),
        description=_(u'Views available to the user on the calendar.'),
        value_type=schema.Choice(
            source=calendar_views
        ),
        default=default_views,
        constraint=select_at_least_one
    )

    form.widget(available_views=CheckBoxFieldWidget)

    selected_view = schema.Choice(
        title=_(u'Selected View'),
        description=_(u'Selected view when opening the calendar.'),
        source=calendar_views,
        default=default_selected_view
    )

    form.widget(selected_view=RadioFieldWidget)

    selected_date = schema.Choice(
        title=_(u'Selected Date'),
        description=_(u'Calendar date shown when opening the calendar.'),
        source=calendar_dates,
        default='current'
    )

    form.widget(selected_date=RadioFieldWidget)

    specific_date = schema.Date(
        title=_(u'Specific Date'),
        required=False
    )

    form.fieldset(
        'defaults',
        label=_(u'Default Allocation Values'),
        fields=(
            'quota', 'partly_available', 'raster', 'approve_manually',
            'reservation_quota_limit'
        )
    )

    form.widget(thank_you_text=WysiwygFieldWidget)
    thank_you_text = schema.Text(
        title=_(u"Thank you text"),
        description=_(
            u"Shown to the user when the reservation has been completed. "
            u"A good place to inform the user general information about the "
            u"reservation he just completed."
        ),
        required=False
    )

    formsets = schema.List(
        title=_(u'Formsets'),
        description=_(
            u'Subforms that need to be filled out to make a reservation. '
            u'Forms can currently only be created by a site-administrator.'
        ),
        value_type=schema.Choice(
            source=form_interfaces,
        ),
        required=False
    )

    form.widget(formsets=CheckBoxFieldWidget)

    @invariant
    def isValidFirstLastHour(Resource):
        in_valid_range = lambda h: 0 <= h and h <= 24
        first_hour, last_hour = Resource.first_hour, Resource.last_hour

        if not in_valid_range(first_hour):
            raise Invalid(_(u'Invalid first hour'))

        if not in_valid_range(last_hour):
            raise Invalid(_(u'Invalid last hour'))

        if last_hour <= first_hour:
            raise Invalid(
                _(u'First hour must be smaller than last hour')
            )

    @invariant
    def isValidCalendarDate(Resource):
        if Resource.selected_date == 'specific' and not Resource.specific_date:
            raise Invalid(
                _(u"You chose to 'Always show a specific date' but you did "
                  u"not specify a specific date")
            )

    @invariant
    def isValidSelectedView(Resource):
        if Resource.selected_view not in Resource.available_views:
            raise Invalid(
                _(u'The selected view must be one of the available views.')
            )


class IResource(IResourceBase):
    pass


class IAllocation(IResourceAllocationDefaults):
    """ An reservable time-slot within a calendar. """

    id = schema.Int(
        title=_(u'Id'),
        default=-1,
        required=False,
    )

    group = schema.Text(
        title=_(u'Recurrence'),
        default=u'',
        required=False
    )

    timeframes = schema.Text(
        title=_(u'Timeframes'),
        default=u'',
        required=False
    )

    start_time = schema.Time(
        title=_(u'Start'),
        description=_(
            u'Allocations may start every 5 minutes if the allocation '
            u'is not partly available. If it is partly available the start '
            u'time may be every x minute where x equals the given raster.'
        )
    )

    end_time = schema.Time(
        title=_(u'End'),
        description=_(
            u'Allocations may end every 5 minutes if the allocation '
            u'is not partly available. If it is partly available the start '
            u'time may be every x minute where x equals the given raster. '
            u'The minimum length of an allocation is also either 5 minutes '
            u'or whatever the value of the raster is.'
        )
    )

    whole_day = schema.Bool(
        title=_(u'Whole Day'),
        description=_(
            u'The allocation spans the whole day.'
        ),
        required=False,
        default=False
    )

    recurring = schema.Choice(
        title=_(u'Recurrence'),
        vocabulary=recurrence,
        default=False
    )

    day = schema.Date(
        title=_(u'Day'),
    )

    recurrence_start = schema.Date(
        title=_(u'From'),
    )

    recurrence_end = schema.Date(
        title=_(u'Until')
    )

    days = schema.List(
        title=_(u'Days'),
        value_type=schema.Choice(vocabulary=days),
        required=False
    )

    separately = schema.Bool(
        title=_(u'Separately reservable'),
        description=_(
            u'If checked parts of the recurrance may be reserved. '
            u'If not checked the recurrance must be reserved as a whole.'
        ),
        required=False,
        default=False
    )

    @invariant
    def isValidRange(Allocation):
        start, end = utils.get_date_range(
            Allocation.day,
            Allocation.start_time, Allocation.end_time
        )

        if not Allocation.whole_day and abs((end - start).seconds // 60) < 5:
            raise Invalid(_(u'The allocation must be at least 5 minutes long'))

    @invariant
    def isValidOption(Allocation):
        if Allocation.recurring:
            if Allocation.partly_available and not Allocation.separately:
                raise Invalid(_(
                    u'Partly available allocations can only be reserved '
                    u'separately'
                ))


class ITimeframe(form.Schema):
    """ A timespan which is either visible or hidden. """

    title = schema.TextLine(
        title=_(u'Name')
    )

    start = schema.Date(
        title=_(u'Start')
    )

    end = schema.Date(
        title=_(u'End')
    )

    @invariant
    def isValidDateRange(Timeframe):
        if Timeframe.start > Timeframe.end:
            raise Invalid(_(u'End date before start date'))

template_variables = _(
    u'May contain the following template variables:<br>'
    u'%(resource)s - title of the resource<br>'
    u'%(dates)s - list of dates reserved<br>'
    u'%(quota)s - reservation quota<br>'
    u'%(reservation_mail)s - email of reservee<br>'
    u'%(data)s - formdata associated with the reservation<br>'
    u'%(reservation_link)s - link to the reservation view<br>'
    u'%(approval_link)s - link to the approval view<br>'
    u'%(denial_link)s - link to the denial view<br>'
    u'%(cancel_link)s - link to the cancel view'
)

template_revoke_variables = template_variables + _(
    u'%(reason)s - reason for revocation<br>'
)

reservations_template_variables = _(
    u'May contain the following template variable:<br>'
    u'%(reservations)s - list of reservations'
)


class IEmailTemplate(form.Schema):
    """ An email template used for custom email messages """

    language = schema.Choice(
        title=_(u'Language'),
        source=plone_languages
    )

    reservation_made_subject = schema.TextLine(
        title=_(u'Email Subject for Reservation Autoapproved'),
        description=_(u'Sent to <b>managers</b> when a reservation is '
                      u'automatically approved. '
                      u'May contain the template variables listed below.'),
        default=templates['reservation_made'].get_subject('en')
    )

    reservation_made_content = schema.Text(
        title=_(u'Email Text for Reservation Autoapproved'),
        description=template_variables,
        default=templates['reservation_made'].get_body('en')
    )

    reservation_pending_subject = schema.TextLine(
        title=_(u'Email Subject for Reservation Pending'),
        description=_(
            u'Sent to <b>managers</b> when a new pending reservation is made. '
            u'May contain the template variables listed below.'
        ),
        default=templates['reservation_pending'].get_subject('en')
    )

    reservation_pending_content = schema.Text(
        title=_(u'Email Text for Reservation Pending'),
        description=template_variables,
        default=templates['reservation_pending'].get_body('en')
    )

    reservation_received_subject = schema.TextLine(
        title=_(u'Email Subject for Received Reservations'),
        description=_(
            u'Sent to <b>users</b> when a new reservation is made wheter they '
            u'still need to be approved or not. By default it cointains the '
            u'reservations made with an indication of the approval status. '
            u'May contain the template variables listed below.'
        ),
        default=templates['reservation_received'].get_subject('en')
    )

    reservation_received_content = schema.Text(
        title=_(u'Email Text for Received Reservations'),
        description=reservations_template_variables,
        default=templates['reservation_received'].get_body('en')
    )

    reservation_approved_subject = schema.TextLine(
        title=_(u'Email Subject for Approved Reservations'),
        description=_(u'Sent to <b>users</b> when a reservation is approved. '
                      u'May contain the template variables listed below.'),
        default=templates['reservation_approved'].get_subject('en')
    )

    reservation_approved_content = schema.Text(
        title=_(u'Email Text for Approved Reservations'),
        description=template_variables,
        default=templates['reservation_approved'].get_body('en')
    )

    reservation_denied_subject = schema.TextLine(
        title=_(u'Email Subject for Denied Reservations'),
        description=_(u'Sent to <b>users</b> when a reservation is denied. '
                      u'May contain the template variables listed below.'),
        default=templates['reservation_denied'].get_subject('en')
    )

    reservation_denied_content = schema.Text(
        title=_(u'Email Text for Denied Reservations'),
        description=template_variables,
        default=templates['reservation_denied'].get_body('en')
    )

    reservation_revoked_subject = schema.TextLine(
        title=_(u'Email Subject for Revoked Reservations'),
        description=_(u'Sent to <b>users</b> when a reservation is revoked. '
                      u'May contain the template variables listed below.'),
        default=templates['reservation_revoked'].get_subject('en')
    )

    reservation_revoked_content = schema.Text(
        title=_(u'Email Text for Revoked Reservations'),
        description=template_revoke_variables,
        default=templates['reservation_revoked'].get_body('en')
    )


def get_default_language(adapter):
    return utils.get_current_site_language()

DefaultLanguage = widget.ComputedWidgetAttribute(
    get_default_language, field=IEmailTemplate['language']
)


class IReservation(Interface):
    """ A reservation of an allocation (may be pending or approved). """

    id = schema.Int(
        title=_(u'Id'),
        default=-1,
        required=False
    )

    day = schema.Date(
        title=_(u'Day'),
        required=False
    )

    start_time = schema.Time(
        title=_(u'Start'),
        required=False
    )

    end_time = schema.Time(
        title=_(u'End'),
        required=False
    )

    quota = schema.Int(
        title=_(u'Number of Reservations'),
        required=False,
        default=1
    )

    email = Email(
        title=_(u'Email'),
        required=True
    )


class IReservationIdForm(Interface):
    """ Describes a form with a hidden reservation-id field. Use with
    seantis.reservation.reserve.ReservationIdForm. """

    reservation = schema.Text(
        title=_(u'Reservation'),
        required=False
    )


class IGroupReservation(Interface):
    """ A reservation of an allocation group. """

    group = schema.Text(
        title=_(u'Recurrence'),
        required=False
    )

    quota = schema.Int(
        title=_(u'Number of Reservations'),
        required=False,
        default=1
    )

    email = Email(
        title=_(u'Email'),
        required=True
    )


class IRevokeReservation(IReservationIdForm):
    """ For the reservation revocation form. """

    send_email = schema.Bool(
        title=_(u"Send Email"),
        description=_(
            u"Send an email to the reservee informing him of the revocation"
        ),
        default=True
    )

    reason = schema.Text(
        title=_(u'Reason'),
        description=_(
            u"Optional reason for the revocation. Sent to the reservee. "
            u"e.g. 'Your reservation has to be cancelled because the lecturer "
            u"is ill'."
        ),
        required=False
    )


class IResourceViewedEvent(Interface):
    """ Event triggered when a seantis.reservation resource is viewed. Pretty
    useful if you need a hook which is guaranteed to be triggered on a plone
    site where seantis.reservation is active.

    """

    context = Attribute("The IResourceBase context object")


class IReservationBaseEvent(Interface):
    """ Base Interface for reservation events (not actually fired). """

    reservation = Attribute("The reservation record associated with the event")
    language = Attribute("The language of the site or current request")


class IReservationMadeEvent(IReservationBaseEvent):
    """ Event triggered when a reservation is made (autoapproved or
        added to the pending reservation list).

    """


class IReservationApprovedEvent(IReservationBaseEvent):
    """ Event triggered when a reservation is approved. """


class IReservationDeniedEvent(IReservationBaseEvent):
    """ Event triggered when a reservation is denied. """


class IReservationRevokedEvent(IReservationBaseEvent):
    """ Event triggered when a reservation is revoked. """

    reason = Attribute("""
        Optional reason for the revocation given by manager. The reason is
        given in the language of the one writing it as the language of the
        reservee is unknown at this point. In the future we might have to
        store said language on the reservation.
    """)


class IReservationsConfirmedEvent(Interface):
    """ Event triggered when the user confirms a list of reservations
    (i.e. submits them).

    Note how this is not a IReservationBaseEvent because it contains
    _multiple_ reservations, not just one.

    """
    reservations = Attribute("The list of reservations the user confirmed")
    language = Attribute("language of the site or current request")



from seantis.reservation import (
    form,
    utils
)

from seantis.reservation import _


class GeneralReportParametersMixin(
    form.ReservationDataView,
    form.ResourceParameterView
):

    titles = {}

    def resource_title(self, uuid):
        if not uuid in self.titles:
            self.titles[uuid] = utils.get_resource_title(self.resources[uuid])
        
        return self.titles[uuid]
    
    @property
    def statuses(self):
        return (
            ('pending', _(u'Pending')),
            ('approved', _(u'Approved')),
        )

    @property
    def hidden_statuses(self):
        return utils.pack(self.request.get('hide_status', []))

    @property
    def reservations(self):
        return utils.pack(self.request.get('reservations', []))

    @property
    def sorted_resources(self):
        objs = self.resources

        sortkey = lambda item: self.resource_title(item[0])
        return utils.OrderedDict(sorted(objs.items(), key=sortkey))

    @property
    def hidden_resources(self):
        return utils.pack(self.request.get('hide_resource', []))

    @property
    def show_details(self):
        # show_details is the query parameter as hiding is the default
        return True if self.request.get('show_details') else False

    @utils.cached_property
    def data_macro_path(self):
        resource = self.resources[self.uuids[0]]
        url = resource.absolute_url() + \
            '/@@reservations/macros/reservation_data'

        return url.replace(self.context.absolute_url(), 'context')

    def show_status(self, status):
        return status not in self.hidden_statuses

    def show_resource(self, uuid):
        return uuid not in self.hidden_resources

    def build_url(self, extra_parameters=None):

        url = '{}/{}'.format(self.context.absolute_url(), self.__name__)
        params = []

        if self.show_details:
            params.append(('show_details', '1'))

        for status in self.hidden_statuses:
            params.append(('hide_status', status))

        for resource in self.hidden_resources:
            params.append(('hide_resource', resource))

        for uuid in self.uuids:
            params.append(('uuid', uuid))

        if extra_parameters:
            params.extend(extra_parameters)

        pair = lambda item: '='.join(item)
        return '{}?{}'.format(url, '&'.join(map(pair, params)))



# -*- coding: utf-8 -*-

import sys

from logging import getLogger
log = getLogger('seantis.reservation')

from datetime import date, datetime, timedelta

from byteplay import *

import seantis.reservation

from seantis.reservation import utils
from seantis.reservation.error import CustomReservationError

allowed_op_codes = set([
    POP_TOP, ROT_TWO, ROT_THREE, ROT_FOUR, DUP_TOP, DUP_TOPX,
    POP_BLOCK, SETUP_LOOP, BUILD_LIST, BUILD_MAP, BUILD_TUPLE,
    LOAD_CONST, RETURN_VALUE, STORE_SUBSCR, STORE_MAP,
    UNARY_POSITIVE, UNARY_NEGATIVE, UNARY_NOT,
    UNARY_INVERT, BINARY_POWER, BINARY_MULTIPLY,
    BINARY_DIVIDE, BINARY_FLOOR_DIVIDE, BINARY_TRUE_DIVIDE,
    BINARY_MODULO, BINARY_ADD, BINARY_SUBTRACT, BINARY_SUBSCR,
    BINARY_LSHIFT, BINARY_RSHIFT, BINARY_AND, BINARY_XOR,
    BINARY_OR, INPLACE_ADD, INPLACE_SUBTRACT, INPLACE_MULTIPLY,
    INPLACE_DIVIDE, INPLACE_POWER,
    INPLACE_LSHIFT, INPLACE_RSHIFT, INPLACE_AND,
    INPLACE_XOR, INPLACE_OR, LOAD_NAME, CALL_FUNCTION, COMPARE_OP, LOAD_ATTR,
    STORE_NAME, GET_ITER, FOR_ITER, LIST_APPEND, DELETE_NAME,
    JUMP_FORWARD, POP_JUMP_IF_TRUE, POP_JUMP_IF_FALSE, JUMP_ABSOLUTE,
    JUMP_IF_TRUE_OR_POP, JUMP_IF_FALSE_OR_POP,
    MAKE_FUNCTION, SLICE_0, SLICE_1, SLICE_2, SLICE_3,
    JUMP_IF_FALSE_OR_POP, JUMP_IF_TRUE_OR_POP, POP_JUMP_IF_FALSE,
    POP_JUMP_IF_TRUE, SETUP_EXCEPT, END_FINALLY
])


def validate_expression(expression, mode='eval'):
    try:
        code = compile(expression, '<dynamic>', mode)
    except (SyntaxError, TypeError):
        raise

    used_codes = set(
        i[0] for i in Code.from_code(code).code if isinstance(i[0], Opcode)
    )

    for opcode in used_codes:
        if opcode not in allowed_op_codes:

            raise ValueError("{} is not an allowed opcode".format(opcode))

    return code


def evaluate_expression(expression, globals_=None, locals_=None, mode='eval'):
    globals_ = globals_ if globals_ is not None else {}
    locals_ = locals_ if locals_ is not None else {}

    globals_.update(
        __builtins__={
            'True': True,
            'False': False,
            'None': None,
            'str': str,
            'globals': locals,
            'locals': locals,
            'bool': bool,
            'dict': dict,
            'list': list,
            'tuple': tuple,
            'map': map,
            'abs': abs,
            'min': min,
            'max': max,
            'reduce': reduce,
            'filter': filter,
            'round': round,
            'len': len,
            'set': set
        }
    )

    code = validate_expression(expression, mode=mode)
    return eval(code, globals_, locals_)


def run_pre_reserve_script(context, start, end, data, locals_=None):
    """ Runs the python script found in the seantis reservation settings
    under a restricted environment with only a selected number of functions
    and variables available.

    The script is able to validate the reservation before it is done,
    giving the ability to implement very specific validation requirements
    of customers with custom fields.

    Obviously this script can't be testet and is insecure by nature. So it
    should be used sparingly and with caution.

    The available methods and variables are purposely undocumented on the
    user interface, because it's meant for developers, not users.
    """
    script = seantis.reservation.settings.get('pre_reservation_script')
    script = isinstance(script, basestring) and unicode(script) or u''
    script = script.strip()

    if not script:
        return

    errors = []

    locals_ = locals_ if locals_ is not None else {}
    locals_.update({
        'path': '/'.join(context.getPhysicalPath()),
        'is_group_reservation': not(start and end),
        'start': start,
        'end': end,
        'date': date,
        'datetime': datetime,
        'timedelta': timedelta,
        'error': errors.append,
        'exit': sys.exit,
        'log': log,
        'formset_available': lambda formset: formset in data
    })
    locals_.update(utils.additional_data_objects(data))

    try:
        evaluate_expression(script, locals_=locals_, mode='exec')
    except SystemExit:
        return
    if errors:
        raise CustomReservationError(errors[0])



from plone.app.testing import TEST_USER_ID

from seantis.reservation import settings
from seantis.reservation.session import serialized
from seantis.reservation.tests import IntegrationTestCase
from seantis.reservation.mail import (
    get_managers_by_context, get_manager_emails
)


class MailTestCase(IntegrationTestCase):

    @serialized
    def test_manager_discovery(self):
        self.login_manager()

        resource = self.create_resource()

        # nobody's defined in the hierarchy, in which case anybody with the
        # reservation role get's the email
        self.set_test_user_roles(['Manager', 'Reservation-Manager'])
        self.assertEqual(get_managers_by_context(resource), [TEST_USER_ID])

        # if we assign two separate managers to the resource we should
        # only get those
        self.assign_reservation_manager('ted@example.com', resource)
        self.assign_reservation_manager('brad@example.com', resource)

        self.assertEqual(
            sorted(get_managers_by_context(resource)),
            sorted(['ted', 'brad'])
        )

        settings.set('send_email_to_managers', 'by_path')
        self.assertEqual(
            sorted(get_manager_emails(resource)),
            ['brad@example.com', 'ted@example.com']
        )

        settings.set('manager_email', u'manager@example.org')
        settings.set('send_email_to_managers', 'by_address')

        self.assertEqual(
            get_manager_emails(resource),
            ['manager@example.org']
        )

        settings.set('send_email_to_managers', 'never')
        self.assertEqual(
            get_manager_emails(resource),
            []
        )



from uuid import UUID

from zope.security import checkPermission
from zope.component import getMultiAdapter

from seantis.reservation.utils import is_uuid, get_resource_by_uuid
from seantis.reservation.utils import string_uuid, real_uuid
from seantis.reservation.timeframe import timeframes_by_context


def for_allocations(context, resources):
    """Returns a function which takes an allocation and returns true if the
    allocation can be exposed.

    resources can be a list of uuids or a list of resource objects

    """

    # get a dictionary with uuids as keys and resources as values
    def get_object(obj):
        if is_uuid(obj):
            return UUID(obj), get_resource_by_uuid(obj)
        else:
            return UUID(obj.uuid()), obj

    resource_objects = dict([get_object(o) for o in resources])

    # get timeframes for each uuid
    timeframes = {}
    for uuid, resource in resource_objects.items():

        # Don't load the timeframes of the resources for which the user has
        # special access to. This way they won't get checked later in
        # 'is_exposed'
        if checkPermission(
                'seantis.reservation.ViewHiddenAllocations', resource):
            timeframes[uuid] = []
        elif resource is None:
            timeframes[uuid] = [None]
        else:
            timeframes[uuid] = timeframes_by_context(resource)

    # returning closure
    def is_exposed(allocation):

        # use the mirror_of as resource-keys of mirrors do not really exist
        # as plone objects
        frames = timeframes[allocation.mirror_of]

        if frames is None or len(frames) == 0:
            return True

        if frames == [None]:
            return False

        # the start date is relevant
        day = allocation.start.date()
        for frame in frames:
            if frame.start <= day and day <= frame.end:
                return frame.visible()

        return False

    return is_exposed


def for_views(context, request):
    """Returns a function which takes a viewname and returns true if the user
    has the right to see the view.

    """

    #gets an instance of the view
    get_view = lambda name: getMultiAdapter((context, request), name=name)

    def is_exposed(viewname):
        view = get_view(viewname)
        assert hasattr(view, 'permission'), "missing permission attribute"

        return checkPermission(view.permission, view)

    return is_exposed


def for_calendar(resource):
    """Returns a function which takes a calendar option and returns true
    if it is enabled.

    """

    option_permissions = {
        'selectable': 'cmf.AddPortalContent',  # drag and drop creation
        'editable': 'cmf.ModifyPortalContent',  # drag and drop resizing
    }

    # right now there's nothing sophisticated to see here
    def is_exposed(option):
        if option in option_permissions:
            return checkPermission(option_permissions[option], resource)
        else:
            return False

    return is_exposed


def for_resources(resources):
    """Returns a function which takes a resource (object or uuid) and
    returns true if it is visible to the current user.

    """

    visible_resources = []

    for r in resources:
        if checkPermission('zope2.View', r):
            visible_resources.append(string_uuid(r))

    def is_exposed(resource):
        return string_uuid(resource) in visible_resources

    return is_exposed


def limit_resources(resources):
    """Given a list of resources or a dictionary with uuid -> resources,
    this function will return the subset of the argument depending on the
    result of for_resource_reservations.is_exposed.

    """

    is_dict = isinstance(resources, dict)
    is_list = isinstance(resources, (list, tuple, set))

    assert is_dict or is_list

    if is_list:
        resdict = dict(((real_uuid(r), r) for r in resources))
    else:
        resdict = resources

    is_exposed = for_resources(resdict.values())

    to_remove = []

    for key, resource in resdict.items():
        if not is_exposed(resource):
            to_remove.append(key)

    for key in to_remove:
        del resdict[key]

    if is_list:
        return resdict.values()
    else:
        return resdict



from five import grok
from zope.interface import Interface

from seantis.reservation.reserve import ReservationUrls
from seantis.reservation.form import ReservationDataView
from seantis.reservation import utils
from seantis.reservation.base import BaseView


class View(BaseView, ReservationUrls, ReservationDataView):
    """A numer of macros for use with seantis.dir.base"""

    grok.context(Interface)
    grok.require('zope2.View')
    grok.name('seantis-reservation-macros')

    template = grok.PageTemplateFile('templates/macros.pt')

    def __getitem__(self, key):
        return self.template._template.macros[key]

    @property
    def utils(self):
        return utils



from setuptools import setup, find_packages
import os

name = "seantis.reservation"
description = "Plone addon to reserve stuff in a calendar."
version = '1.0.24'

zug_require = [
    'ftw.contentmenu',
    'izug.basetheme',
]
teamraum_require = [
    'plonetheme.teamraum'
]
tests_require = [
    'collective.betterbrowser[pyquery]',
    'collective.testcaselayer',
    'plone.app.testing',
]


def get_long_description():
    readme = open('README.rst').read()
    history = open(os.path.join('docs', 'HISTORY.txt')).read()

    # cut the part before the description to avoid repetition on pypi
    readme = readme[readme.index(description) + len(description):]

    return '\n'.join((readme, history))

setup(name=name, version=version, description=description,
      long_description=get_long_description(),
      classifiers=[
          'Framework :: Plone',
          'Framework :: Plone :: 4.3',
          'Intended Audience :: Developers',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Programming Language :: Python',
      ],
      keywords='reservation calendar seantis plone dexterity',
      author='Seantis GmbH',
      author_email='info@seantis.ch',
      url='https://github.com/seantis/seantis.reservation',
      license='GPL v2',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['seantis'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'alembic>=0.5.0',
          'byteplay',
          'collective.autopermission',
          'collective.js.fullcalendar>=1.6.1',
          'collective.js.jqueryui',
          'collective.js.underscore',
          'isodate',
          'mock',
          'ordereddict',
          'plone.api',
          'plone.app.dexterity [grok]',
          'plone.app.referenceablebehavior',
          'plone.app.z3cform>=0.7.6',
          'plone.behavior',
          'plone.directives.form',
          'plone.resourceeditor',
          'plone.uuid>=1.0.2',
          'Plone>=4.3',
          'profilehooks',
          'psycopg2',
          'pytz',
          'setuptools',
          'seantis.plonetools>=0.11',
          'SQLAlchemy>=0.7.3',
          'tablib',
          'xlwt',
          'zope.sqlalchemy',
      ],
      extras_require=dict(
          zug=zug_require,
          tests=tests_require,
          teamraum=teamraum_require
      ),
      entry_points="""
      # -*- Entry points: -*-

      [z3c.autoinclude.plugin]
      target = plone
      """
      )



from sqlalchemy.exc import IntegrityError
from psycopg2.extensions import TransactionRollbackError
from sqlalchemy.orm.exc import NoResultFound

from seantis.reservation import _


class ReservationError(Exception):
    pass


class OverlappingAllocationError(ReservationError):

    def __init__(self, start, end, existing):
        self.start = start
        self.end = end
        self.existing = existing


class AffectedReservationError(ReservationError):

    def __init__(self, existing):
        self.existing = existing


class CustomReservationError(ReservationError):

    def __init__(self, msg):
        self.msg = msg


class AffectedPendingReservationError(AffectedReservationError):
    pass


class AlreadyReservedError(ReservationError):
    pass


class DirtyReadOnlySession(ReservationError):
    pass


class ModifiedReadOnlySession(ReservationError):
    pass


class NotReservableError(ReservationError):
    pass


class ReservationTooLong(ReservationError):
    pass


class ThrottleBlock(ReservationError):
    pass


class ReservationParametersInvalid(ReservationError):
    pass


class InvalidReservationToken(ReservationError):
    pass


class InvalidReservationError(ReservationError):
    pass


class QuotaOverLimit(ReservationError):
    pass


class InvalidQuota(ReservationError):
    pass


class QuotaImpossible(ReservationError):
    pass


class InvalidAllocationError(ReservationError):
    pass


class NoReservationsToConfirm(ReservationError):
    pass


errormap = {

    OverlappingAllocationError:
    _(u'A conflicting allocation exists for the requested time period.'),

    AffectedReservationError:
    _(u'An existing reservation would be affected by the requested change.'),

    AffectedPendingReservationError:
    _(u'A pending reservation would be affected by the requested change.'),

    TransactionRollbackError:
    _(u'The resource is being edited by someone else. Please try again.'),

    NoResultFound:
    _(u'The item does no longer exist.'),

    AlreadyReservedError:
    _(u'The requested period is no longer available.'),

    IntegrityError:
    _(u'Invalid change. Your request may have already been processed '
      u'earlier.'),

    NotReservableError:
    _(u'No reservable slot found.'),

    ReservationTooLong:
    _(u"Reservations can't be made for more than 24 hours at a time"),

    ThrottleBlock:
    _(u'Too many reservations in a short time. Wait for a moment before '
      u'trying again.'),

    ReservationParametersInvalid:
    _(u'The given reservation paramters are invalid.'),

    InvalidReservationToken:
    _(u'The given reservation token is invalid.'),

    InvalidReservationError:
    _(u'The given reservation paramters are invalid.'),

    QuotaOverLimit:
    _(u'The requested number of reservations is higher than allowed.'),

    InvalidQuota:
    _(u'The requested quota is invalid (must be at least one).'),

    QuotaImpossible:
    _(u'The allocation does not have enough spots to possibly satisfy the '
      u'requested number of reservations.'),

    InvalidAllocationError:
    _(u'The resulting allocation would be invalid.'),

    NoReservationsToConfirm:
    _(u'No reservations to confirm.')

}



from datetime import datetime
from uuid import uuid1 as uuid

from sqlalchemy.exc import IntegrityError

from seantis.reservation import Session
from seantis.reservation.tests import IntegrationTestCase
from seantis.reservation.models import Allocation
from seantis.reservation.session import serialized


class TestAllocation(IntegrationTestCase):

    @serialized
    def test_simple_add(self):
        # Test a simple add
        allocation = Allocation(raster=15, resource=uuid())
        allocation.start = datetime(2011, 1, 1, 15)
        allocation.end = datetime(2011, 1, 1, 15, 59)
        allocation.group = str(uuid())
        allocation.mirror_of = allocation.resource

        Session.add(allocation)

        self.assertEqual(Session.query(Allocation).count(), 1)

        # Test failing add
        allocation = Allocation(raster=15)

        Session.add(allocation)
        self.assertRaises(IntegrityError, Session.flush)

    def test_date_functions(self):
        allocation = Allocation(raster=60, resource=uuid())
        allocation.start = datetime(2011, 1, 1, 12, 30)
        allocation.end = datetime(2011, 1, 1, 14, 00)

        self.assertEqual(allocation.start.hour, 12)
        self.assertEqual(allocation.start.minute, 0)

        self.assertEqual(allocation.end.hour, 13)
        self.assertEqual(allocation.end.minute, 59)

        start = datetime(2011, 1, 1, 11, 00)
        end = datetime(2011, 1, 1, 12, 05)

        self.assertTrue(allocation.overlaps(start, end))
        self.assertFalse(allocation.contains(start, end))

        start = datetime(2011, 1, 1, 13, 00)
        end = datetime(2011, 1, 1, 15, 00)

        self.assertTrue(allocation.overlaps(start, end))
        self.assertFalse(allocation.contains(start, end))

    def test_whole_day(self):
        allocation = Allocation(raster=15, resource=uuid())

        allocation.start = datetime(2013, 1, 1, 0, 0)
        allocation.end = datetime(2013, 1, 2, 0, 0)

        self.assertTrue(allocation.whole_day)

        allocation.start = datetime(2013, 1, 1, 0, 0)
        allocation.end = datetime(2013, 1, 1, 23, 59, 59, 999999)

        self.assertTrue(allocation.whole_day)

        allocation.start = datetime(2013, 1, 1, 0, 0)
        allocation.end = datetime(2013, 1, 2, 23, 59, 59, 999999)

        self.assertTrue(allocation.whole_day)

        allocation.start = datetime(2013, 1, 1, 0, 0)
        allocation.end = datetime(2013, 1, 2, 0, 0)

        self.assertTrue(allocation.whole_day)

        allocation.start = datetime(2013, 1, 1, 15, 0)
        allocation.end = datetime(2013, 1, 1, 0, 0)

        self.assertRaises(AssertionError, lambda: allocation.whole_day)



import uuid
from zope.site.hooks import getSite
from Products.CMFCore.utils import getToolByName


session_key = lambda name: '%s:reservation:%s' % (getSite().id, name)

# users need to get a predictable uuid valid for each plone site
# ideally, a plone site would provide us with a uuid as a namespace
# from which to generate further uuids, but it doesn't, so we need
# to do that ourselves. The root namespace is used to get a plone site
# specific uuid which is then used as a namespace for the users.
root_namespace = uuid.UUID('7ad36f77-a10c-4b62-857c-a9a4ff7222a0')


def generate_session_id(context):
    """ Generates a new session id. """

    membership = getToolByName(context, 'portal_membership')

    # anonymous users get random uuids
    if membership.isAnonymousUser():
        return uuid.uuid4()

    # logged in users get ids which are predictable for each plone site
    namespace = uuid.uuid5(root_namespace, str(getSite().id))
    return uuid.uuid5(namespace,
                      str(membership.getAuthenticatedMember().getId()))


def get_session(context, key):
    """Gets the key from the session."""
    session_manager = getToolByName(context, 'session_data_manager')

    if not session_manager.hasSessionData():
        return None

    session = session_manager.getSessionData()

    if not key in session.keys():
        return None

    return session[key]


def set_session(context, key, value):
    """Stores the given value with the key in the session."""
    session_manager = getToolByName(context, 'session_data_manager')
    session = session_manager.getSessionData()
    session[key] = value


def get_session_id(context):
    """Returns the current session id (models/reservation/session_id), creating
    it first if necessary.

    """
    skey = session_key('session_id')
    session_id = get_session(context, skey)

    if session_id is None:
        session_id = generate_session_id(context)
        set_session(context, skey, session_id)

    return session_id


def get_email(context):
    return get_session(context, session_key('email')) or None


def set_email(context, email):
    set_session(context, session_key('email'), email)


def get_additional_data(context):
    return get_session(context, session_key('additional_data'))


def set_additional_data(context, data):
    set_session(context, session_key('additional_data'), data)



import transaction

from App.config import getConfiguration, setConfiguration
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import IntegrationTesting
from plone.app.testing import FunctionalTesting
from plone.app.testing import applyProfile
from plone.app.testing import quickInstallProduct
from plone.testing import z2
from OFS.Folder import Folder
from Testing import ZopeTestCase
from zope.configuration import xmlconfig

try:
    from seantis.reservation import test_database
except ImportError:
    from seantis.reservation.utils import ConfigurationError
    msg = 'No test database configured in seantis.reservation.test_database.'
    raise ConfigurationError(msg)


class SqlLayer(PloneSandboxLayer):

    default_bases = (PLONE_FIXTURE,)

    class Session(dict):
        def set(self, key, value):
            self[key] = value

    @property
    def dsn(self):
        if not self.get('dsn'):
            self['dsn'] = test_database.testdsn

        return self['dsn']

    def init_config(self):
        config = getConfiguration()
        if not hasattr(config, 'product_config'):
            config.product_config = {}

        config.product_config['seantis.reservation'] = dict(dsn=self.dsn)

        setConfiguration(config)

    def setUpZope(self, app, configurationContext):

        # Set up sessioning objects
        app.REQUEST['SESSION'] = self.Session()
        ZopeTestCase.utils.setupCoreSessions(app)

        self.init_config()

        import seantis.reservation
        xmlconfig.file(
            'configure.zcml',
            seantis.reservation,
            context=configurationContext
        )
        self.loadZCML(package=seantis.reservation)

    def setUpPloneSite(self, portal):

        quickInstallProduct(portal, 'plone.app.dexterity')
        quickInstallProduct(portal, 'seantis.reservation')
        applyProfile(portal, 'seantis.reservation:default')

    def tearDownZope(self, app):
        z2.uninstallProduct(app, 'seantis.reservation')

SQL_FIXTURE = SqlLayer()

SQL_INTEGRATION_TESTING = IntegrationTesting(
    bases=(SQL_FIXTURE, ),
    name="SqlLayer:Integration"
)

SQL_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(SQL_FIXTURE, ),
    name="SqlLayer:Functional"
)



from itertools import groupby

import logging
log = logging.getLogger('seantis.reservation')

from five import grok

from email.MIMEText import MIMEText
from email.Header import Header
from email.Utils import parseaddr, formataddr

from plone.dexterity.content import Item
from plone.directives import dexterity
from plone.memoize import view
from Products.CMFCore.interfaces import IFolderish
from Products.CMFCore.utils import getToolByName
from z3c.form import button
from zope.schema import getFields

from seantis.reservation.form import ReservationDataView
from seantis.reservation.reserve import ReservationUrls
from seantis.reservation.base import BaseViewlet
from seantis.reservation.interfaces import IReservationsConfirmedEvent
from seantis.reservation.interfaces import IReservationApprovedEvent
from seantis.reservation.interfaces import IReservationDeniedEvent
from seantis.reservation.interfaces import IReservationRevokedEvent
from seantis.reservation.interfaces import OverviewletManager
from seantis.reservation.interfaces import IEmailTemplate
from seantis.reservation.interfaces import ISeantisReservationSpecific
from seantis.reservation.mail_templates import templates
from seantis.reservation import utils
from seantis.reservation import settings
from seantis.reservation import _


@grok.subscribe(IReservationsConfirmedEvent)
def on_reservations_confirmed(event):

    # send one mail to the reservee
    if settings.get('send_email_to_reservees'):
        send_reservations_confirmed(event.reservations, event.language)

    # send many mails to the admins
    if settings.get('send_email_to_managers') != 'never':
        for reservation in event.reservations:

            if reservation.autoapprovable:
                send_reservation_mail(
                    reservation,
                    'reservation_made', event.language, to_managers=True
                )
            else:
                send_reservation_mail(
                    reservation,
                    'reservation_pending', event.language, to_managers=True
                )


@grok.subscribe(IReservationApprovedEvent)
def on_reservation_approved(event):
    if not settings.get('send_email_to_reservees'):
        return
    if not event.reservation.autoapprovable:
        send_reservation_mail(
            event.reservation, 'reservation_approved', event.language
        )


@grok.subscribe(IReservationDeniedEvent)
def on_reservation_denied(event):
    if not settings.get('send_email_to_reservees'):
        return
    if not event.reservation.autoapprovable:
        send_reservation_mail(
            event.reservation, 'reservation_denied', event.language
        )


@grok.subscribe(IReservationRevokedEvent)
def on_reservation_revoked(event):
    if not settings.get('send_email_to_reservees'):
        return

    if not event.send_email:
        return

    send_reservation_mail(
        event.reservation, 'reservation_revoked', event.language,
        to_managers=False, revocation_reason=event.reason
    )


class EmailTemplate(Item):
    def get_title(self):
        return _(u'Email Template') + u' ' + \
            utils.native_language_name(self.language)

    def set_title(self, title):
        pass

    title = property(get_title, set_title)


def get_managers_by_context(context):
    managers = context.users_with_local_role('Reservation-Manager')

    if managers:
        return managers

    if not hasattr(context, 'portal_type'):
        return []

    # if we arrive at the top level we just notify whoever got the reservation
    # manager role on the site
    if context.portal_type == 'Plone Site':
        return [
            m.id for m in
            getToolByName(context, 'portal_membership').listMembers()
            if m.has_role('Reservation-Manager')
        ]

    return get_managers_by_context(context.aq_inner.aq_parent)


def get_manager_emails_by_context(context):
    managers = get_managers_by_context(context)

    if not managers:
        return []

    acl = utils.getToolByName(context, 'acl_users')
    groups = acl.source_groups.getGroupIds()

    # remove the groups and replace them with users
    userids = []
    for man in managers:
        if man in groups:
            userids.extend(acl.source_groups.getGroupById(man).getMemberIds())
        else:
            userids.append(man)

    userids = set(userids)

    # go through the users and get their emails
    emails = []
    for uid in userids:
        user = acl.getUserById(uid)

        if user:
            emails.append(user.getProperty('email'))
        else:
            log.warn('The manager with the id %s does not exist' % uid)

    return emails


def get_manager_emails(context):
    if settings.get('send_email_to_managers') == 'by_path':
        return get_manager_emails_by_context(context)
    elif settings.get('send_email_to_managers') == 'by_address':
        return [settings.get('manager_email')]
    else:
        return []


def get_email_content(context, email_type, language):
    user_templates = utils.portal_type_by_context(
        context, portal_type='seantis.reservation.emailtemplate'
    )

    for t in user_templates:
        if t.language != language:
            continue

        subject = getattr(t, email_type + '_subject')
        body = getattr(t, email_type + '_content')

        return subject, body

    return templates[email_type].get(language)


def load_resources(reservations):
    resources = dict()

    for resource in (r.resource for r in reservations):
        if resource in resources:
            continue

        resources[resource] = utils.get_resource_by_uuid(resource).getObject()

    return resources


def may_send_mail(resource, mail, intended_for_admin):
    if intended_for_admin:
        return True

    if mail.recipient in get_manager_emails(resource):
        return False

    return True


def send_reservations_confirmed(reservations, language):

    sender = utils.get_site_email_sender()

    if not sender:
        log.warn('Cannot send email as no sender is configured')
        return

    # load resources
    resources = load_resources(reservations)

    # send reservations grouped by reservee email
    groupkey = lambda r: r.email
    by_recipient = groupby(sorted(reservations, key=groupkey), key=groupkey)

    for recipient, grouped_reservations in by_recipient:

        lines = []

        for reservation in grouped_reservations:

            resource = resources[reservation.resource]

            prefix = '' if reservation.autoapprovable else '* '
            title_prefix = '{}x '.format(reservation.quota)
            lines.append(
                prefix + utils.get_resource_title(resource, title_prefix)
            )

            for start, end in reservation.timespans():
                lines.append(utils.display_date(start, end))

            lines.append('')

        # differs between resources
        subject, body = get_email_content(
            resource, 'reservation_received', language
        )

        mail = ReservationMail(
            resource, reservation,
            sender=sender,
            recipient=recipient,
            subject=subject,
            body=body,
            reservations=lines[:-1]
        )

        if may_send_mail(resource, mail, intended_for_admin=False):
            send_mail(resource, mail)


def send_reservation_mail(reservation, email_type, language,
                          to_managers=False, revocation_reason=u''):

    resource = utils.get_resource_by_uuid(reservation.resource)

    # the resource doesn't currently exist in testing so we quietly
    # exit. This should be changed => #TODO
    if not resource:
        log.warn('Cannot send email as the resource does not exist')
        return

    sender = utils.get_site_email_sender()

    if not sender:
        log.warn('Cannot send email as no sender is configured')
        return

    resource = resource.getObject()

    if to_managers:
        recipients = get_manager_emails(resource)
        if not recipients:
            log.warn("Couldn't find a manager to send an email to")
            return
    else:
        recipients = [reservation.email]

    subject, body = get_email_content(resource, email_type, language)

    for recipient in recipients:
        mail = ReservationMail(
            resource, reservation,
            sender=sender,
            recipient=recipient,
            subject=subject,
            body=body,
            revocation_reason=revocation_reason
        )

        if may_send_mail(resource, mail, intended_for_admin=to_managers):
            send_mail(resource, mail)


def send_mail(context, mail):
    context.MailHost.send(mail.as_string(), immediate=False)


class ReservationMail(ReservationDataView, ReservationUrls):

    sender = u''
    recipient = u''
    subject = u''
    body = u''
    reservations = u''
    revocation_reason = u''

    def __init__(self, resource, reservation, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)

        # get information for the body/subject string

        p = dict()
        is_needed = lambda key: key in self.subject or key in self.body

        # title of the resource
        if is_needed('resource'):
            p['resource'] = utils.get_resource_title(resource)

        # reservation email
        if is_needed('reservation_mail'):
            p['reservation_mail'] = reservation.email

        # a list of reservations
        if is_needed('reservations'):
            p['reservations'] = '\n'.join(self.reservations)

        # a list of dates
        if is_needed('dates'):
            lines = []
            dates = sorted(reservation.timespans(), key=lambda i: i[0])
            for start, end in dates:
                lines.append(utils.display_date(start, end))

            p['dates'] = '\n'.join(lines)

        # reservation quota
        if is_needed('quota'):
            p['quota'] = reservation.quota

        # tabbed reservation data
        if is_needed('data'):
            data = reservation.data
            lines = []
            for key in self.sort_reservation_data(data):
                interface = data[key]

                lines.append(interface['desc'])
                sorted_values = self.sort_reservation_data_values(
                    interface['values']
                )

                for value in sorted_values:
                    lines.append(
                        '\t' + value['desc'] + ': ' +
                        unicode(self.display_reservation_data(value['value']))
                    )

            p['data'] = '\n'.join(lines)

        if is_needed('reservation_link'):
            p['reservation_link'] = self.show_all_url(
                reservation.token, resource
            )

        # approval link
        if is_needed('approval_link'):
            p['approval_link'] = self.approve_all_url(
                reservation.token, resource
            )

        # denial link
        if is_needed('denial_link'):
            p['denial_link'] = self.deny_all_url(reservation.token, resource)

        # cancel link
        if is_needed('cancel_link'):
            p['cancel_link'] = self.revoke_all_url(reservation.token, resource)

        # revocation reason
        if is_needed('reason'):
            p['reason'] = self.revocation_reason

        self.parameters = p

    def as_string(self):
        subject = self.subject % self.parameters
        body = self.body % self.parameters
        mail = create_email(self.sender, self.recipient, subject, body)
        return mail.as_string()


def create_email(sender, recipient, subject, body):
    """Create an email message.

    All arguments should be Unicode strings (plain ASCII works as well).

    Only the real name part of sender and recipient addresses may contain
    non-ASCII characters.

    The charset of the email will be the UTF-8.
    """

    header_charset = 'UTF-8'
    body_charset = 'UTF-8'

    body.encode(body_charset)

    # Split real name (which is optional) and email address parts
    sender_name, sender_addr = parseaddr(sender)
    recipient_name, recipient_addr = parseaddr(recipient)

    # We must always pass Unicode strings to Header, otherwise it will
    # use RFC 2047 encoding even on plain ASCII strings.
    sender_name = str(Header(unicode(sender_name), header_charset))
    recipient_name = str(Header(unicode(recipient_name), header_charset))

    # Make sure email addresses do not contain non-ASCII characters
    sender_addr = sender_addr.encode('ascii')
    recipient_addr = recipient_addr.encode('ascii')

    # Create the message ('plain' stands for Content-Type: text/plain)
    msg = MIMEText(body.encode(body_charset), 'plain', body_charset)
    msg['From'] = formataddr((sender_name, sender_addr))
    msg['To'] = formataddr((recipient_name, recipient_addr))
    msg['Subject'] = Header(unicode(subject), header_charset)

    return msg


# The following is quite similar to the timeframe.py stuff
# surely this can be merged somewhat => #TODO

def validate_template(context, request, data):
    if context.portal_type == 'seantis.reservation.emailtemplate':
        folder = context.aq_inner.aq_parent
    else:
        folder = context

    templates = utils.portal_type_in_context(
        folder, portal_type='seantis.reservation.emailtemplate'
    )

    duplicate = False

    for template in templates:
        if template.id == context.id:
            continue

        if template.getObject().title == context.title:
            duplicate = True
            break

    if duplicate:
        msg = utils.translate(
            context, request,
            _(u"There's already an Email template in the same folder for the "
              u"same language")
        )
        utils.form_error(msg)


class TemplateAddForm(dexterity.AddForm):

    permission = 'cmf.AddPortalContent'

    grok.context(IEmailTemplate)
    grok.require(permission)
    grok.layer(ISeantisReservationSpecific)

    grok.name('seantis.reservation.emailtemplate')

    def update(self, **kwargs):
        super(TemplateAddForm, self).update(**kwargs)
        self.use_translated_emails_as_default()

    def get_field_map(self, suffix):
        return dict(
            (field.replace(suffix, ''), field)
            for field in getFields(self.schema) if field.endswith(suffix)
        )

    def apply_field_map(self, mapping, get_template_value):
        for template, field in mapping.items():
            value = get_template_value(template)
            if value:
                self.widgets[field].value = value

    def use_translated_emails_as_default(self):
        language = utils.get_current_language(self.context, self.request)[:2]

        subjects = self.get_field_map('_subject')
        contents = self.get_field_map('_content')

        self.apply_field_map(
            subjects, lambda tpl: templates[tpl].get_subject(language)
        )
        self.apply_field_map(
            contents, lambda tpl: templates[tpl].get_body(language)
        )

    @button.buttonAndHandler(_('Save'), name='save')
    def handleAdd(self, action):
        data, errors = self.extractData()
        validate_template(self.context, self.request, data)
        dexterity.AddForm.handleAdd(self, action)


class TemplateEditForm(dexterity.EditForm):

    permission = 'cmf.ModifyPortalContent'

    grok.context(IEmailTemplate)
    grok.require(permission)
    grok.layer(ISeantisReservationSpecific)

    @button.buttonAndHandler(_(u'Save'), name='save')
    def handleApply(self, action):
        data, errors = self.extractData()
        validate_template(self.context, self.request, data)
        dexterity.EditForm.handleApply(self, action)


class TemplatesViewlet(BaseViewlet):

    permission = 'cmf.ModifyPortalContent'

    grok.context(IFolderish)
    grok.require(permission)

    grok.name('seantis.reservation.mailviewlet')
    grok.viewletmanager(OverviewletManager)

    grok.order(4)

    _template = grok.PageTemplateFile('templates/email_templates.pt')

    @view.memoize
    def templates(self):
        templates = [t.getObject() for t in utils.portal_type_in_context(
            self.context, portal_type='seantis.reservation.emailtemplate'
        )]
        return sorted(templates, key=lambda t: t.title)

    def links(self, template=None):

        # global links
        if not template:
            baseurl = self.context.absolute_url()
            return [(
                _(u'Add email template'),
                baseurl + '/++add++seantis.reservation.emailtemplate'
            )]

        # template specific links
        links = []

        baseurl = template.absolute_url()
        links.append((_(u'Edit'), baseurl + '/edit'))
        links.append((_(u'Delete'), baseurl + '/delete_confirmation'))

        return links

    def render(self, **kwargs):
        return self._template.render(self)



import random

from datetime import datetime, timedelta

from App.config import getConfiguration
from five import grok
from zope.interface import Interface

from plone.dexterity.utils import createContentInContainer

from seantis.reservation.models import Allocation
from seantis.reservation.session import serialized, Session
from seantis.reservation.error import (
    OverlappingAllocationError,
    ReservationError
)
from seantis.reservation.raster import VALID_RASTER_VALUES
from seantis.reservation.base import BaseView

import transaction


class DataGeneratorView(BaseView):

    permission = 'cmf.ManagePortal'
    grok.require(permission)

    grok.context(Interface)
    grok.name('generate')

    template = grok.PageTemplateFile('templates/datagenerator.pt')

    @property
    def may_run(self):
        return getConfiguration().debug_mode

    @property
    def start(self):
        start = self.request.get('start', None)
        if start:
            return datetime.strptime(start, '%d.%m.%Y')
        else:
            return None

    @property
    def end(self):
        end = self.request.get('end', None)
        if end:
            return datetime.strptime(end, '%d.%m.%Y')
        else:
            return None

    @property
    def with_reservations(self):
        return bool(self.request.get('with_reservations', None))

    @property
    def min_duration(self):
        return int(self.request.get('min_duration', 30))

    @property
    def first_hour(self):
        return int(self.request.get('first_hour', 8))

    @property
    def last_hour(self):
        return int(self.request.get('last_hour', 18))

    def create_resource(self):
        resource = createContentInContainer(
            self.context, 'seantis.reservation.resource',
            title=u'random @ ' + datetime.today().strftime('%d.%m.%Y %H:%M')
        )
        resource.first_hour = self.first_hour
        resource.last_hour = self.last_hour
        return resource

    @serialized
    def generate_allocations(self, resource=None, start=None, end=None):

        today = datetime.today()

        resource = resource or self.create_resource()
        start = start or datetime(today.year, 1, 1)
        end = end or (start + timedelta(days=365))

        days = []
        for day in xrange(0, (end - start).days, 1):
            days.append(start + timedelta(days=day))

        scheduler = resource.scheduler()

        for day in days:

            for timespan in self.random_timespans(resource, day):

                quota = random.randrange(1, 1000)

                print 'a @', timespan[0], timespan[1]

                try:
                    scheduler.allocate(
                        (timespan[0], timespan[1]),
                        raster=timespan[2],
                        partly_available=bool(random.randrange(0, 2)),
                        grouped=False,
                        quota=quota,
                        approve_manually=bool(random.randrange(0, 2))
                    )
                except OverlappingAllocationError:
                    pass

            # we must commit regularly or the postgres serial session
            # must track so many queries it goes to the barn and puts itself
            # down

            transaction.commit()

        if self.with_reservations:
            self.generate_reservations(resource, start, end)

    @serialized
    def generate_reservations(self, resource, start, end):
        query = Session.query(Allocation)
        query = query.filter(Allocation._start >= start)
        query = query.filter(Allocation._end <= end)
        query = query.filter(Allocation.mirror_of == resource.string_uuid())
        query = query.order_by(Allocation._start)

        email = 'generated@example.com'
        scheduler = resource.scheduler()

        allocations = query.all()
        Session.expunge_all()

        for allocation in allocations:

            if allocation.partly_available:
                start = allocation.start
                total = (allocation.end - allocation.start).seconds / 60

                if total > allocation.raster:
                    start_minute = random.randrange(
                        0, total - allocation.raster, allocation.raster
                    )
                    end_minute = random.randrange(
                        start_minute + allocation.raster, total,
                        allocation.raster
                    )
                else:
                    start_minute = 0
                    end_minute = total

                start = start + timedelta(start_minute * 60)
                end = start + timedelta(end_minute * 60)
            else:
                start, end = allocation.start, allocation.display_end

            if not allocation.approve_manually:
                limit = allocation.quota

            for i in xrange(0, random.randrange(0, limit + 1)):
                try:
                    print 'r @', start, end
                    token = scheduler.reserve(email, dates=(start, end))
                    if not allocation.approve_manually:
                        scheduler.approve_reservation(token)
                except ReservationError:
                    break

            Session.expire_on_commit = False
            transaction.commit()

    def random_raster(self):
        return random.choice(VALID_RASTER_VALUES)

    def random_timespans(self, resource, day):

        day = day or datetime.today()

        min_minute = resource.first_hour * 60
        max_minute = resource.last_hour * 60

        timespans = []

        base = datetime(day.year, day.month, day.day)

        while True:
            raster = self.random_raster()
            offset = max(raster, self.min_duration)

            if max_minute - min_minute <= offset:
                break

            start_minute = random.randrange(
                min_minute, max_minute - offset, raster
            )
            end_minute = random.randrange(
                start_minute + offset, max_minute, raster
            )

            start = base + timedelta(seconds=start_minute * 60)
            end = base + timedelta(seconds=end_minute * 60)

            timespans.append((start, end, raster))

            min_minute = end_minute

        return timespans

    def update(self, *args, **kwargs):
        super(DataGeneratorView, self).update(*args, **kwargs)

        if self.may_run and self.request.get('generate_data'):
            self.generate_allocations(start=self.start, end=self.end)
            print "done!"



import logging
logger = logging.getLogger('seantis.reservation')

from zope import schema
from zope.interface import Interface, Invalid, invariant
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm
from z3c.form.browser.radio import RadioFieldWidget

from plone import api
from plone.directives import form
from plone.z3cform import layout
from plone.app.registry.browser.controlpanel import RegistryEditForm
from plone.app.registry.browser.controlpanel import ControlPanelFormWrapper

from Products.Five.browser.pagetemplatefile import ZopeTwoPageTemplateFile

from seantis.plonetools.schemafields import validate_email

from seantis.reservation import _
from seantis.reservation.restricted_eval import validate_expression


def valid_expression(expression):
    try:
        validate_expression(expression, mode='exec')
    except Exception, e:
        raise Invalid(str(e))
    return True


class ISeantisReservationSettings(Interface):

    throttle_minutes = schema.Int(
        title=_(u"Reservation Throttling"),
        description=_(
            u'The number of minutes a user needs to wait between '
            u'reservations, use 0 if no throttling should occur. '
            u'Users with the \'Unthrottled Reservations\' permission '
            u'are excempt from this rule (Reservation-Managers and '
            u'Reservation-Assistants by default).'
        )
    )

    form.widget(send_email_to_managers=RadioFieldWidget)
    send_email_to_managers = schema.Choice(
        title=_(u"Email Notifications for Managers"),
        description=_(
            u'Either send an email to the manager(s) responsible for the '
            u'resource or use a single recipient address for all resources.'
        ),
        source=SimpleVocabulary(
            [
                SimpleTerm(
                    value='never',
                    title=_(u'Do not send manager emails')
                ),
                SimpleTerm(
                    value='by_path',
                    title=_(u'Send to the responsible manager(s)')
                ),
                SimpleTerm(
                    value='by_address',
                    title=_(u'Always send to the following address:')
                )
            ]
        ),
        default='by_path'
    )

    # XXX
    # the plone registry is a bit weird in how it handles fields, it will
    # not work with seantis.plonetools' Email field - so we use TextLine with
    # a valid_email constraint.
    manager_email = schema.TextLine(
        title=_(u"Email Address for Manager Emails"),
        required=False,
        default=u'',
        constraint=validate_email
    )

    send_email_to_reservees = schema.Bool(
        title=_(u"Email Notifications for Reservees"),
        description=_(
            u'Send emails about made, approved and denied reservations '
            u'to the user that made the reservation.'
        )
    )

    pre_reservation_script = schema.Text(
        title=_(u"Pre-Reservation Script"),
        description=_(
            u'Run custom validation code for reservations. This is for '
            u'advanced users only and may disable the reservations process. '
            u'For documentation study the source code at Github.'
        ),
        required=False,
        constraint=valid_expression
    )

    available_threshold = schema.Int(
        title=_(u'Available Threshold (%)'),
        description=_(
            u'Allocations with an availability above or equal to this value '
            u'are considered available and are shown green'
        ),
        default=75
    )

    partly_available_threshold = schema.Int(
        title=_(u'Partly Available Threshold (%)'),
        description=_(
            u'Allocations with an availability above or equal to this value '
            u'but below the available threshold are considered partly '
            u'available and are shown orange. Allocations with an '
            u'availability below this value are considered unavailable and '
            u'are shown red'
        ),
        default=1
    )

    @invariant
    def isValidThreshold(Allocation):

        thresholds = [
            Allocation.available_threshold,
            Allocation.partly_available_threshold,
        ]

        if thresholds[0] == thresholds[1]:
            raise Invalid(
                _(u'Thresholds must have differing values')
            )

        if not (thresholds[0] > thresholds[1]):
            raise Invalid(
                _(
                    u'The available threshold must have a larger value than '
                    u'the partly available threshold'
                )
            )

        for threshold in thresholds:
            if not (0 <= threshold and threshold <= 100):
                raise Invalid(
                    _(
                        u'Thresholds must have values between 0 and 100'
                    )
                )

    @invariant
    def filled_manager_email(settings):
        if settings.send_email_to_managers == 'by_address':
            if not (settings.manager_email or u'').strip():
                raise Invalid(_(
                    u'Please specify an address to send the manager emails to'
                ))


def get(name):
    prefix = ISeantisReservationSettings.__identifier__
    return api.portal.get_registry_record('.'.join((prefix, name)))


def set(name, value):
    prefix = ISeantisReservationSettings.__identifier__
    return api.portal.set_registry_record('.'.join((prefix, name)), value)


class SeantisReservationSettingsPanelForm(RegistryEditForm):
    schema = ISeantisReservationSettings
    label = _(u"Seantis Reservation Control Panel")

    template = ZopeTwoPageTemplateFile('templates/controlpanel.pt')


SeantisReservationControlPanelView = layout.wrap_form(
    SeantisReservationSettingsPanelForm, ControlPanelFormWrapper
)
